# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha1ServerStorageVersion(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'api_server_id': 'str',
        'decodable_versions': 'list[str]',
        'encoding_version': 'str',
        'served_versions': 'list[str]'
    }

    attribute_map = {
        'api_server_id': 'apiServerID',
        'decodable_versions': 'decodableVersions',
        'encoding_version': 'encodingVersion',
        'served_versions': 'servedVersions'
    }

    def __init__(self, api_server_id=None, decodable_versions=None, encoding_version=None, served_versions=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1ServerStorageVersion - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._api_server_id = None
        self._decodable_versions = None
        self._encoding_version = None
        self._served_versions = None
        self.discriminator = None

        self.api_server_id = api_server_id
        self.decodable_versions = decodable_versions
        self.encoding_version = encoding_version
        if served_versions is not None:
            self.served_versions = served_versions

    @property
    def api_server_id(self):
        """Gets the api_server_id of this V1alpha1ServerStorageVersion.  # noqa: E501

        apiServerID is the ID of the reporting API server.  # noqa: E501

        :return: The api_server_id of this V1alpha1ServerStorageVersion.  # noqa: E501
        :rtype: str
        """
        return self._api_server_id

    @api_server_id.setter
    def api_server_id(self, api_server_id):
        """Sets the api_server_id of this V1alpha1ServerStorageVersion.

        apiServerID is the ID of the reporting API server.  # noqa: E501

        :param api_server_id: The api_server_id of this V1alpha1ServerStorageVersion.  # noqa: E501
        :type api_server_id: str
        """
        if self.local_vars_configuration.client_side_validation and api_server_id is None:  # noqa: E501
            raise ValueError("Invalid value for `api_server_id`, must not be `None`")  # noqa: E501

        self._api_server_id = api_server_id

    @property
    def decodable_versions(self):
        """Gets the decodable_versions of this V1alpha1ServerStorageVersion.  # noqa: E501

        decodableVersions are the encoding versions the API server can handle to decode. The API server can decode objects encoded in these versions. The encodingVersion must be included in the decodableVersions.  # noqa: E501

        :return: The decodable_versions of this V1alpha1ServerStorageVersion.  # noqa: E501
        :rtype: list[str]
        """
        return self._decodable_versions

    @decodable_versions.setter
    def decodable_versions(self, decodable_versions):
        """Sets the decodable_versions of this V1alpha1ServerStorageVersion.

        decodableVersions are the encoding versions the API server can handle to decode. The API server can decode objects encoded in these versions. The encodingVersion must be included in the decodableVersions.  # noqa: E501

        :param decodable_versions: The decodable_versions of this V1alpha1ServerStorageVersion.  # noqa: E501
        :type decodable_versions: list[str]
        """
        if self.local_vars_configuration.client_side_validation and decodable_versions is None:  # noqa: E501
            raise ValueError("Invalid value for `decodable_versions`, must not be `None`")  # noqa: E501

        self._decodable_versions = decodable_versions

    @property
    def encoding_version(self):
        """Gets the encoding_version of this V1alpha1ServerStorageVersion.  # noqa: E501

        encodingVersion the API server encodes the object to when persisting it in the backend (e.g., etcd).  # noqa: E501

        :return: The encoding_version of this V1alpha1ServerStorageVersion.  # noqa: E501
        :rtype: str
        """
        return self._encoding_version

    @encoding_version.setter
    def encoding_version(self, encoding_version):
        """Sets the encoding_version of this V1alpha1ServerStorageVersion.

        encodingVersion the API server encodes the object to when persisting it in the backend (e.g., etcd).  # noqa: E501

        :param encoding_version: The encoding_version of this V1alpha1ServerStorageVersion.  # noqa: E501
        :type encoding_version: str
        """
        if self.local_vars_configuration.client_side_validation and encoding_version is None:  # noqa: E501
            raise ValueError("Invalid value for `encoding_version`, must not be `None`")  # noqa: E501

        self._encoding_version = encoding_version

    @property
    def served_versions(self):
        """Gets the served_versions of this V1alpha1ServerStorageVersion.  # noqa: E501

        servedVersions lists all versions the API server can serve. DecodableVersions must include all ServedVersions.  # noqa: E501

        :return: The served_versions of this V1alpha1ServerStorageVersion.  # noqa: E501
        :rtype: list[str]
        """
        return self._served_versions

    @served_versions.setter
    def served_versions(self, served_versions):
        """Sets the served_versions of this V1alpha1ServerStorageVersion.

        servedVersions lists all versions the API server can serve. DecodableVersions must include all ServedVersions.  # noqa: E501

        :param served_versions: The served_versions of this V1alpha1ServerStorageVersion.  # noqa: E501
        :type served_versions: list[str]
        """

        self._served_versions = served_versions

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1ServerStorageVersion):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1ServerStorageVersion):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class DiscoveryV1EndpointPort(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'app_protocol': 'str',
        'name': 'str',
        'port': 'int',
        'protocol': 'str'
    }

    attribute_map = {
        'app_protocol': 'appProtocol',
        'name': 'name',
        'port': 'port',
        'protocol': 'protocol'
    }

    def __init__(self, app_protocol=None, name=None, port=None, protocol=None, local_vars_configuration=None):  # noqa: E501
        """DiscoveryV1EndpointPort - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._app_protocol = None
        self._name = None
        self._port = None
        self._protocol = None
        self.discriminator = None

        if app_protocol is not None:
            self.app_protocol = app_protocol
        if name is not None:
            self.name = name
        if port is not None:
            self.port = port
        if protocol is not None:
            self.protocol = protocol

    @property
    def app_protocol(self):
        """Gets the app_protocol of this DiscoveryV1EndpointPort.  # noqa: E501

        The application protocol for this port. This is used as a hint for implementations to offer richer behavior for protocols that they understand. This field follows standard Kubernetes label syntax. Valid values are either:  * Un-prefixed protocol names - reserved for IANA standard service names (as per RFC-6335 and https://www.iana.org/assignments/service-names).  * Kubernetes-defined prefixed names:   * 'kubernetes.io/h2c' - HTTP/2 prior knowledge over cleartext as described in https://www.rfc-editor.org/rfc/rfc9113.html#name-starting-http-2-with-prior-   * 'kubernetes.io/ws'  - WebSocket over cleartext as described in https://www.rfc-editor.org/rfc/rfc6455   * 'kubernetes.io/wss' - WebSocket over TLS as described in https://www.rfc-editor.org/rfc/rfc6455  * Other protocols should use implementation-defined prefixed names such as mycompany.com/my-custom-protocol.  # noqa: E501

        :return: The app_protocol of this DiscoveryV1EndpointPort.  # noqa: E501
        :rtype: str
        """
        return self._app_protocol

    @app_protocol.setter
    def app_protocol(self, app_protocol):
        """Sets the app_protocol of this DiscoveryV1EndpointPort.

        The application protocol for this port. This is used as a hint for implementations to offer richer behavior for protocols that they understand. This field follows standard Kubernetes label syntax. Valid values are either:  * Un-prefixed protocol names - reserved for IANA standard service names (as per RFC-6335 and https://www.iana.org/assignments/service-names).  * Kubernetes-defined prefixed names:   * 'kubernetes.io/h2c' - HTTP/2 prior knowledge over cleartext as described in https://www.rfc-editor.org/rfc/rfc9113.html#name-starting-http-2-with-prior-   * 'kubernetes.io/ws'  - WebSocket over cleartext as described in https://www.rfc-editor.org/rfc/rfc6455   * 'kubernetes.io/wss' - WebSocket over TLS as described in https://www.rfc-editor.org/rfc/rfc6455  * Other protocols should use implementation-defined prefixed names such as mycompany.com/my-custom-protocol.  # noqa: E501

        :param app_protocol: The app_protocol of this DiscoveryV1EndpointPort.  # noqa: E501
        :type app_protocol: str
        """

        self._app_protocol = app_protocol

    @property
    def name(self):
        """Gets the name of this DiscoveryV1EndpointPort.  # noqa: E501

        name represents the name of this port. All ports in an EndpointSlice must have a unique name. If the EndpointSlice is derived from a Kubernetes service, this corresponds to the Service.ports[].name. Name must either be an empty string or pass DNS_LABEL validation: * must be no more than 63 characters long. * must consist of lower case alphanumeric characters or '-'. * must start and end with an alphanumeric character. Default is empty string.  # noqa: E501

        :return: The name of this DiscoveryV1EndpointPort.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DiscoveryV1EndpointPort.

        name represents the name of this port. All ports in an EndpointSlice must have a unique name. If the EndpointSlice is derived from a Kubernetes service, this corresponds to the Service.ports[].name. Name must either be an empty string or pass DNS_LABEL validation: * must be no more than 63 characters long. * must consist of lower case alphanumeric characters or '-'. * must start and end with an alphanumeric character. Default is empty string.  # noqa: E501

        :param name: The name of this DiscoveryV1EndpointPort.  # noqa: E501
        :type name: str
        """

        self._name = name

    @property
    def port(self):
        """Gets the port of this DiscoveryV1EndpointPort.  # noqa: E501

        port represents the port number of the endpoint. If the EndpointSlice is derived from a Kubernetes service, this must be set to the service's target port. EndpointSlices used for other purposes may have a nil port.  # noqa: E501

        :return: The port of this DiscoveryV1EndpointPort.  # noqa: E501
        :rtype: int
        """
        return self._port

    @port.setter
    def port(self, port):
        """Sets the port of this DiscoveryV1EndpointPort.

        port represents the port number of the endpoint. If the EndpointSlice is derived from a Kubernetes service, this must be set to the service's target port. EndpointSlices used for other purposes may have a nil port.  # noqa: E501

        :param port: The port of this DiscoveryV1EndpointPort.  # noqa: E501
        :type port: int
        """

        self._port = port

    @property
    def protocol(self):
        """Gets the protocol of this DiscoveryV1EndpointPort.  # noqa: E501

        protocol represents the IP protocol for this port. Must be UDP, TCP, or SCTP. Default is TCP.  # noqa: E501

        :return: The protocol of this DiscoveryV1EndpointPort.  # noqa: E501
        :rtype: str
        """
        return self._protocol

    @protocol.setter
    def protocol(self, protocol):
        """Sets the protocol of this DiscoveryV1EndpointPort.

        protocol represents the IP protocol for this port. Must be UDP, TCP, or SCTP. Default is TCP.  # noqa: E501

        :param protocol: The protocol of this DiscoveryV1EndpointPort.  # noqa: E501
        :type protocol: str
        """

        self._protocol = protocol

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DiscoveryV1EndpointPort):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DiscoveryV1EndpointPort):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1PodFailurePolicyRule(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'action': 'str',
        'on_exit_codes': 'V1PodFailurePolicyOnExitCodesRequirement',
        'on_pod_conditions': 'list[V1PodFailurePolicyOnPodConditionsPattern]'
    }

    attribute_map = {
        'action': 'action',
        'on_exit_codes': 'onExitCodes',
        'on_pod_conditions': 'onPodConditions'
    }

    def __init__(self, action=None, on_exit_codes=None, on_pod_conditions=None, local_vars_configuration=None):  # noqa: E501
        """V1PodFailurePolicyRule - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._action = None
        self._on_exit_codes = None
        self._on_pod_conditions = None
        self.discriminator = None

        self.action = action
        if on_exit_codes is not None:
            self.on_exit_codes = on_exit_codes
        if on_pod_conditions is not None:
            self.on_pod_conditions = on_pod_conditions

    @property
    def action(self):
        """Gets the action of this V1PodFailurePolicyRule.  # noqa: E501

        Specifies the action taken on a pod failure when the requirements are satisfied. Possible values are:  - FailJob: indicates that the pod's job is marked as Failed and all   running pods are terminated. - FailIndex: indicates that the pod's index is marked as Failed and will   not be restarted. - Ignore: indicates that the counter towards the .backoffLimit is not   incremented and a replacement pod is created. - Count: indicates that the pod is handled in the default way - the   counter towards the .backoffLimit is incremented. Additional values are considered to be added in the future. Clients should react to an unknown action by skipping the rule.  # noqa: E501

        :return: The action of this V1PodFailurePolicyRule.  # noqa: E501
        :rtype: str
        """
        return self._action

    @action.setter
    def action(self, action):
        """Sets the action of this V1PodFailurePolicyRule.

        Specifies the action taken on a pod failure when the requirements are satisfied. Possible values are:  - FailJob: indicates that the pod's job is marked as Failed and all   running pods are terminated. - FailIndex: indicates that the pod's index is marked as Failed and will   not be restarted. - Ignore: indicates that the counter towards the .backoffLimit is not   incremented and a replacement pod is created. - Count: indicates that the pod is handled in the default way - the   counter towards the .backoffLimit is incremented. Additional values are considered to be added in the future. Clients should react to an unknown action by skipping the rule.  # noqa: E501

        :param action: The action of this V1PodFailurePolicyRule.  # noqa: E501
        :type action: str
        """
        if self.local_vars_configuration.client_side_validation and action is None:  # noqa: E501
            raise ValueError("Invalid value for `action`, must not be `None`")  # noqa: E501

        self._action = action

    @property
    def on_exit_codes(self):
        """Gets the on_exit_codes of this V1PodFailurePolicyRule.  # noqa: E501


        :return: The on_exit_codes of this V1PodFailurePolicyRule.  # noqa: E501
        :rtype: V1PodFailurePolicyOnExitCodesRequirement
        """
        return self._on_exit_codes

    @on_exit_codes.setter
    def on_exit_codes(self, on_exit_codes):
        """Sets the on_exit_codes of this V1PodFailurePolicyRule.


        :param on_exit_codes: The on_exit_codes of this V1PodFailurePolicyRule.  # noqa: E501
        :type on_exit_codes: V1PodFailurePolicyOnExitCodesRequirement
        """

        self._on_exit_codes = on_exit_codes

    @property
    def on_pod_conditions(self):
        """Gets the on_pod_conditions of this V1PodFailurePolicyRule.  # noqa: E501

        Represents the requirement on the pod conditions. The requirement is represented as a list of pod condition patterns. The requirement is satisfied if at least one pattern matches an actual pod condition. At most 20 elements are allowed.  # noqa: E501

        :return: The on_pod_conditions of this V1PodFailurePolicyRule.  # noqa: E501
        :rtype: list[V1PodFailurePolicyOnPodConditionsPattern]
        """
        return self._on_pod_conditions

    @on_pod_conditions.setter
    def on_pod_conditions(self, on_pod_conditions):
        """Sets the on_pod_conditions of this V1PodFailurePolicyRule.

        Represents the requirement on the pod conditions. The requirement is represented as a list of pod condition patterns. The requirement is satisfied if at least one pattern matches an actual pod condition. At most 20 elements are allowed.  # noqa: E501

        :param on_pod_conditions: The on_pod_conditions of this V1PodFailurePolicyRule.  # noqa: E501
        :type on_pod_conditions: list[V1PodFailurePolicyOnPodConditionsPattern]
        """

        self._on_pod_conditions = on_pod_conditions

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1PodFailurePolicyRule):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1PodFailurePolicyRule):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1ResourcePolicyRule(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'api_groups': 'list[str]',
        'cluster_scope': 'bool',
        'namespaces': 'list[str]',
        'resources': 'list[str]',
        'verbs': 'list[str]'
    }

    attribute_map = {
        'api_groups': 'apiGroups',
        'cluster_scope': 'clusterScope',
        'namespaces': 'namespaces',
        'resources': 'resources',
        'verbs': 'verbs'
    }

    def __init__(self, api_groups=None, cluster_scope=None, namespaces=None, resources=None, verbs=None, local_vars_configuration=None):  # noqa: E501
        """V1ResourcePolicyRule - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._api_groups = None
        self._cluster_scope = None
        self._namespaces = None
        self._resources = None
        self._verbs = None
        self.discriminator = None

        self.api_groups = api_groups
        if cluster_scope is not None:
            self.cluster_scope = cluster_scope
        if namespaces is not None:
            self.namespaces = namespaces
        self.resources = resources
        self.verbs = verbs

    @property
    def api_groups(self):
        """Gets the api_groups of this V1ResourcePolicyRule.  # noqa: E501

        `apiGroups` is a list of matching API groups and may not be empty. \"*\" matches all API groups and, if present, must be the only entry. Required.  # noqa: E501

        :return: The api_groups of this V1ResourcePolicyRule.  # noqa: E501
        :rtype: list[str]
        """
        return self._api_groups

    @api_groups.setter
    def api_groups(self, api_groups):
        """Sets the api_groups of this V1ResourcePolicyRule.

        `apiGroups` is a list of matching API groups and may not be empty. \"*\" matches all API groups and, if present, must be the only entry. Required.  # noqa: E501

        :param api_groups: The api_groups of this V1ResourcePolicyRule.  # noqa: E501
        :type api_groups: list[str]
        """
        if self.local_vars_configuration.client_side_validation and api_groups is None:  # noqa: E501
            raise ValueError("Invalid value for `api_groups`, must not be `None`")  # noqa: E501

        self._api_groups = api_groups

    @property
    def cluster_scope(self):
        """Gets the cluster_scope of this V1ResourcePolicyRule.  # noqa: E501

        `clusterScope` indicates whether to match requests that do not specify a namespace (which happens either because the resource is not namespaced or the request targets all namespaces). If this field is omitted or false then the `namespaces` field must contain a non-empty list.  # noqa: E501

        :return: The cluster_scope of this V1ResourcePolicyRule.  # noqa: E501
        :rtype: bool
        """
        return self._cluster_scope

    @cluster_scope.setter
    def cluster_scope(self, cluster_scope):
        """Sets the cluster_scope of this V1ResourcePolicyRule.

        `clusterScope` indicates whether to match requests that do not specify a namespace (which happens either because the resource is not namespaced or the request targets all namespaces). If this field is omitted or false then the `namespaces` field must contain a non-empty list.  # noqa: E501

        :param cluster_scope: The cluster_scope of this V1ResourcePolicyRule.  # noqa: E501
        :type cluster_scope: bool
        """

        self._cluster_scope = cluster_scope

    @property
    def namespaces(self):
        """Gets the namespaces of this V1ResourcePolicyRule.  # noqa: E501

        `namespaces` is a list of target namespaces that restricts matches.  A request that specifies a target namespace matches only if either (a) this list contains that target namespace or (b) this list contains \"*\".  Note that \"*\" matches any specified namespace but does not match a request that _does not specify_ a namespace (see the `clusterScope` field for that). This list may be empty, but only if `clusterScope` is true.  # noqa: E501

        :return: The namespaces of this V1ResourcePolicyRule.  # noqa: E501
        :rtype: list[str]
        """
        return self._namespaces

    @namespaces.setter
    def namespaces(self, namespaces):
        """Sets the namespaces of this V1ResourcePolicyRule.

        `namespaces` is a list of target namespaces that restricts matches.  A request that specifies a target namespace matches only if either (a) this list contains that target namespace or (b) this list contains \"*\".  Note that \"*\" matches any specified namespace but does not match a request that _does not specify_ a namespace (see the `clusterScope` field for that). This list may be empty, but only if `clusterScope` is true.  # noqa: E501

        :param namespaces: The namespaces of this V1ResourcePolicyRule.  # noqa: E501
        :type namespaces: list[str]
        """

        self._namespaces = namespaces

    @property
    def resources(self):
        """Gets the resources of this V1ResourcePolicyRule.  # noqa: E501

        `resources` is a list of matching resources (i.e., lowercase and plural) with, if desired, subresource.  For example, [ \"services\", \"nodes/status\" ].  This list may not be empty. \"*\" matches all resources and, if present, must be the only entry. Required.  # noqa: E501

        :return: The resources of this V1ResourcePolicyRule.  # noqa: E501
        :rtype: list[str]
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """Sets the resources of this V1ResourcePolicyRule.

        `resources` is a list of matching resources (i.e., lowercase and plural) with, if desired, subresource.  For example, [ \"services\", \"nodes/status\" ].  This list may not be empty. \"*\" matches all resources and, if present, must be the only entry. Required.  # noqa: E501

        :param resources: The resources of this V1ResourcePolicyRule.  # noqa: E501
        :type resources: list[str]
        """
        if self.local_vars_configuration.client_side_validation and resources is None:  # noqa: E501
            raise ValueError("Invalid value for `resources`, must not be `None`")  # noqa: E501

        self._resources = resources

    @property
    def verbs(self):
        """Gets the verbs of this V1ResourcePolicyRule.  # noqa: E501

        `verbs` is a list of matching verbs and may not be empty. \"*\" matches all verbs and, if present, must be the only entry. Required.  # noqa: E501

        :return: The verbs of this V1ResourcePolicyRule.  # noqa: E501
        :rtype: list[str]
        """
        return self._verbs

    @verbs.setter
    def verbs(self, verbs):
        """Sets the verbs of this V1ResourcePolicyRule.

        `verbs` is a list of matching verbs and may not be empty. \"*\" matches all verbs and, if present, must be the only entry. Required.  # noqa: E501

        :param verbs: The verbs of this V1ResourcePolicyRule.  # noqa: E501
        :type verbs: list[str]
        """
        if self.local_vars_configuration.client_side_validation and verbs is None:  # noqa: E501
            raise ValueError("Invalid value for `verbs`, must not be `None`")  # noqa: E501

        self._verbs = verbs

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ResourcePolicyRule):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ResourcePolicyRule):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha3PoolStatus(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'allocated_devices': 'int',
        'available_devices': 'int',
        'driver': 'str',
        'generation': 'int',
        'node_name': 'str',
        'pool_name': 'str',
        'resource_slice_count': 'int',
        'total_devices': 'int',
        'unavailable_devices': 'int',
        'validation_error': 'str'
    }

    attribute_map = {
        'allocated_devices': 'allocatedDevices',
        'available_devices': 'availableDevices',
        'driver': 'driver',
        'generation': 'generation',
        'node_name': 'nodeName',
        'pool_name': 'poolName',
        'resource_slice_count': 'resourceSliceCount',
        'total_devices': 'totalDevices',
        'unavailable_devices': 'unavailableDevices',
        'validation_error': 'validationError'
    }

    def __init__(self, allocated_devices=None, available_devices=None, driver=None, generation=None, node_name=None, pool_name=None, resource_slice_count=None, total_devices=None, unavailable_devices=None, validation_error=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha3PoolStatus - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._allocated_devices = None
        self._available_devices = None
        self._driver = None
        self._generation = None
        self._node_name = None
        self._pool_name = None
        self._resource_slice_count = None
        self._total_devices = None
        self._unavailable_devices = None
        self._validation_error = None
        self.discriminator = None

        if allocated_devices is not None:
            self.allocated_devices = allocated_devices
        if available_devices is not None:
            self.available_devices = available_devices
        self.driver = driver
        self.generation = generation
        if node_name is not None:
            self.node_name = node_name
        self.pool_name = pool_name
        if resource_slice_count is not None:
            self.resource_slice_count = resource_slice_count
        if total_devices is not None:
            self.total_devices = total_devices
        if unavailable_devices is not None:
            self.unavailable_devices = unavailable_devices
        if validation_error is not None:
            self.validation_error = validation_error

    @property
    def allocated_devices(self):
        """Gets the allocated_devices of this V1alpha3PoolStatus.  # noqa: E501

        AllocatedDevices is the number of devices currently allocated to claims. A value of 0 means no devices are allocated. May be unset when validationError is set.  # noqa: E501

        :return: The allocated_devices of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._allocated_devices

    @allocated_devices.setter
    def allocated_devices(self, allocated_devices):
        """Sets the allocated_devices of this V1alpha3PoolStatus.

        AllocatedDevices is the number of devices currently allocated to claims. A value of 0 means no devices are allocated. May be unset when validationError is set.  # noqa: E501

        :param allocated_devices: The allocated_devices of this V1alpha3PoolStatus.  # noqa: E501
        :type allocated_devices: int
        """

        self._allocated_devices = allocated_devices

    @property
    def available_devices(self):
        """Gets the available_devices of this V1alpha3PoolStatus.  # noqa: E501

        AvailableDevices is the number of devices available for allocation. This equals TotalDevices - AllocatedDevices - UnavailableDevices. A value of 0 means no devices are currently available. May be unset when validationError is set.  # noqa: E501

        :return: The available_devices of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._available_devices

    @available_devices.setter
    def available_devices(self, available_devices):
        """Sets the available_devices of this V1alpha3PoolStatus.

        AvailableDevices is the number of devices available for allocation. This equals TotalDevices - AllocatedDevices - UnavailableDevices. A value of 0 means no devices are currently available. May be unset when validationError is set.  # noqa: E501

        :param available_devices: The available_devices of this V1alpha3PoolStatus.  # noqa: E501
        :type available_devices: int
        """

        self._available_devices = available_devices

    @property
    def driver(self):
        """Gets the driver of this V1alpha3PoolStatus.  # noqa: E501

        Driver is the DRA driver name for this pool. Must be a DNS subdomain (e.g., \"gpu.example.com\").  # noqa: E501

        :return: The driver of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: str
        """
        return self._driver

    @driver.setter
    def driver(self, driver):
        """Sets the driver of this V1alpha3PoolStatus.

        Driver is the DRA driver name for this pool. Must be a DNS subdomain (e.g., \"gpu.example.com\").  # noqa: E501

        :param driver: The driver of this V1alpha3PoolStatus.  # noqa: E501
        :type driver: str
        """
        if self.local_vars_configuration.client_side_validation and driver is None:  # noqa: E501
            raise ValueError("Invalid value for `driver`, must not be `None`")  # noqa: E501

        self._driver = driver

    @property
    def generation(self):
        """Gets the generation of this V1alpha3PoolStatus.  # noqa: E501

        Generation is the pool generation observed across all ResourceSlices in this pool. Only the latest generation is reported. During a generation rollout, if not all slices at the latest generation have been published, the pool is included with a validationError and device counts unset.  # noqa: E501

        :return: The generation of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._generation

    @generation.setter
    def generation(self, generation):
        """Sets the generation of this V1alpha3PoolStatus.

        Generation is the pool generation observed across all ResourceSlices in this pool. Only the latest generation is reported. During a generation rollout, if not all slices at the latest generation have been published, the pool is included with a validationError and device counts unset.  # noqa: E501

        :param generation: The generation of this V1alpha3PoolStatus.  # noqa: E501
        :type generation: int
        """
        if self.local_vars_configuration.client_side_validation and generation is None:  # noqa: E501
            raise ValueError("Invalid value for `generation`, must not be `None`")  # noqa: E501

        self._generation = generation

    @property
    def node_name(self):
        """Gets the node_name of this V1alpha3PoolStatus.  # noqa: E501

        NodeName is the node this pool is associated with. When omitted, the pool is not associated with a specific node. Must be a valid DNS subdomain name (RFC1123).  # noqa: E501

        :return: The node_name of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: str
        """
        return self._node_name

    @node_name.setter
    def node_name(self, node_name):
        """Sets the node_name of this V1alpha3PoolStatus.

        NodeName is the node this pool is associated with. When omitted, the pool is not associated with a specific node. Must be a valid DNS subdomain name (RFC1123).  # noqa: E501

        :param node_name: The node_name of this V1alpha3PoolStatus.  # noqa: E501
        :type node_name: str
        """

        self._node_name = node_name

    @property
    def pool_name(self):
        """Gets the pool_name of this V1alpha3PoolStatus.  # noqa: E501

        PoolName is the name of the pool. Must be a valid resource pool name (DNS subdomains separated by \"/\").  # noqa: E501

        :return: The pool_name of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: str
        """
        return self._pool_name

    @pool_name.setter
    def pool_name(self, pool_name):
        """Sets the pool_name of this V1alpha3PoolStatus.

        PoolName is the name of the pool. Must be a valid resource pool name (DNS subdomains separated by \"/\").  # noqa: E501

        :param pool_name: The pool_name of this V1alpha3PoolStatus.  # noqa: E501
        :type pool_name: str
        """
        if self.local_vars_configuration.client_side_validation and pool_name is None:  # noqa: E501
            raise ValueError("Invalid value for `pool_name`, must not be `None`")  # noqa: E501

        self._pool_name = pool_name

    @property
    def resource_slice_count(self):
        """Gets the resource_slice_count of this V1alpha3PoolStatus.  # noqa: E501

        ResourceSliceCount is the number of ResourceSlices that make up this pool. May be unset when validationError is set.  # noqa: E501

        :return: The resource_slice_count of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._resource_slice_count

    @resource_slice_count.setter
    def resource_slice_count(self, resource_slice_count):
        """Sets the resource_slice_count of this V1alpha3PoolStatus.

        ResourceSliceCount is the number of ResourceSlices that make up this pool. May be unset when validationError is set.  # noqa: E501

        :param resource_slice_count: The resource_slice_count of this V1alpha3PoolStatus.  # noqa: E501
        :type resource_slice_count: int
        """

        self._resource_slice_count = resource_slice_count

    @property
    def total_devices(self):
        """Gets the total_devices of this V1alpha3PoolStatus.  # noqa: E501

        TotalDevices is the total number of devices in the pool across all slices. A value of 0 means the pool has no devices. May be unset when validationError is set.  # noqa: E501

        :return: The total_devices of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._total_devices

    @total_devices.setter
    def total_devices(self, total_devices):
        """Sets the total_devices of this V1alpha3PoolStatus.

        TotalDevices is the total number of devices in the pool across all slices. A value of 0 means the pool has no devices. May be unset when validationError is set.  # noqa: E501

        :param total_devices: The total_devices of this V1alpha3PoolStatus.  # noqa: E501
        :type total_devices: int
        """

        self._total_devices = total_devices

    @property
    def unavailable_devices(self):
        """Gets the unavailable_devices of this V1alpha3PoolStatus.  # noqa: E501

        UnavailableDevices is the number of devices that are not available due to taints or other conditions, but are not allocated. A value of 0 means all unallocated devices are available. May be unset when validationError is set.  # noqa: E501

        :return: The unavailable_devices of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: int
        """
        return self._unavailable_devices

    @unavailable_devices.setter
    def unavailable_devices(self, unavailable_devices):
        """Sets the unavailable_devices of this V1alpha3PoolStatus.

        UnavailableDevices is the number of devices that are not available due to taints or other conditions, but are not allocated. A value of 0 means all unallocated devices are available. May be unset when validationError is set.  # noqa: E501

        :param unavailable_devices: The unavailable_devices of this V1alpha3PoolStatus.  # noqa: E501
        :type unavailable_devices: int
        """

        self._unavailable_devices = unavailable_devices

    @property
    def validation_error(self):
        """Gets the validation_error of this V1alpha3PoolStatus.  # noqa: E501

        ValidationError is set when the pool's data could not be fully validated (e.g., incomplete slice publication). When set, device count fields and ResourceSliceCount may be unset.  # noqa: E501

        :return: The validation_error of this V1alpha3PoolStatus.  # noqa: E501
        :rtype: str
        """
        return self._validation_error

    @validation_error.setter
    def validation_error(self, validation_error):
        """Sets the validation_error of this V1alpha3PoolStatus.

        ValidationError is set when the pool's data could not be fully validated (e.g., incomplete slice publication). When set, device count fields and ResourceSliceCount may be unset.  # noqa: E501

        :param validation_error: The validation_error of this V1alpha3PoolStatus.  # noqa: E501
        :type validation_error: str
        """

        self._validation_error = validation_error

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha3PoolStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha3PoolStatus):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha3DeviceTaintSelector(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'device': 'str',
        'driver': 'str',
        'pool': 'str'
    }

    attribute_map = {
        'device': 'device',
        'driver': 'driver',
        'pool': 'pool'
    }

    def __init__(self, device=None, driver=None, pool=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha3DeviceTaintSelector - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._device = None
        self._driver = None
        self._pool = None
        self.discriminator = None

        if device is not None:
            self.device = device
        if driver is not None:
            self.driver = driver
        if pool is not None:
            self.pool = pool

    @property
    def device(self):
        """Gets the device of this V1alpha3DeviceTaintSelector.  # noqa: E501

        If device is set, only devices with that name are selected. This field corresponds to slice.spec.devices[].name.  Setting also driver and pool may be required to avoid ambiguity, but is not required.  # noqa: E501

        :return: The device of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :rtype: str
        """
        return self._device

    @device.setter
    def device(self, device):
        """Sets the device of this V1alpha3DeviceTaintSelector.

        If device is set, only devices with that name are selected. This field corresponds to slice.spec.devices[].name.  Setting also driver and pool may be required to avoid ambiguity, but is not required.  # noqa: E501

        :param device: The device of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :type device: str
        """

        self._device = device

    @property
    def driver(self):
        """Gets the driver of this V1alpha3DeviceTaintSelector.  # noqa: E501

        If driver is set, only devices from that driver are selected. This fields corresponds to slice.spec.driver.  # noqa: E501

        :return: The driver of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :rtype: str
        """
        return self._driver

    @driver.setter
    def driver(self, driver):
        """Sets the driver of this V1alpha3DeviceTaintSelector.

        If driver is set, only devices from that driver are selected. This fields corresponds to slice.spec.driver.  # noqa: E501

        :param driver: The driver of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :type driver: str
        """

        self._driver = driver

    @property
    def pool(self):
        """Gets the pool of this V1alpha3DeviceTaintSelector.  # noqa: E501

        If pool is set, only devices in that pool are selected.  Also setting the driver name may be useful to avoid ambiguity when different drivers use the same pool name, but this is not required because selecting pools from different drivers may also be useful, for example when drivers with node-local devices use the node name as their pool name.  # noqa: E501

        :return: The pool of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :rtype: str
        """
        return self._pool

    @pool.setter
    def pool(self, pool):
        """Sets the pool of this V1alpha3DeviceTaintSelector.

        If pool is set, only devices in that pool are selected.  Also setting the driver name may be useful to avoid ambiguity when different drivers use the same pool name, but this is not required because selecting pools from different drivers may also be useful, for example when drivers with node-local devices use the node name as their pool name.  # noqa: E501

        :param pool: The pool of this V1alpha3DeviceTaintSelector.  # noqa: E501
        :type pool: str
        """

        self._pool = pool

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha3DeviceTaintSelector):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha3DeviceTaintSelector):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2026 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Thread-safe in-memory store for the Kubernetes informer."""

import threading


def _meta_namespace_key(obj):
    """Build a lookup key from object metadata.

    Supports both dict-based objects and generated model objects.
    Returns namespace/name for namespaced objects, just name otherwise.
    """
    if isinstance(obj, dict):
        meta = obj.get("metadata") or {}
        ns = meta.get("namespace") or ""
        name = meta.get("name") or ""
    else:
        meta = getattr(obj, "metadata", None)
        if meta is None:
            return ""
        if hasattr(meta, "namespace"):
            ns = getattr(meta, "namespace", None) or ""
            name = getattr(meta, "name", None) or ""
        else:
            ns = meta.get("namespace") or ""
            name = meta.get("name") or ""
    if ns:
        return "{}/{}".format(ns, name)
    return name


class ObjectCache:
    """Thread-safe in-memory mapping of Kubernetes objects.

    The SharedInformer keeps this store synchronised with the API server.
    Consumers can call list() and get_by_key() from any thread safely.
    """

    def __init__(self, key_func=None):
        self._key_func = key_func if key_func is not None else _meta_namespace_key
        self._objects = {}
        self._rlock = threading.RLock()

    # --- mutation helpers (called by SharedInformer) ---

    def _put(self, obj):
        key = self._key_func(obj)
        with self._rlock:
            self._objects[key] = obj

    def _remove(self, obj):
        key = self._key_func(obj)
        with self._rlock:
            self._objects.pop(key, None)

    def _replace_all(self, objects):
        rebuilt = {self._key_func(o): o for o in objects}
        with self._rlock:
            self._objects = rebuilt

    # --- public read API ---

    def list(self):
        """Return a snapshot list of all cached objects."""
        with self._rlock:
            return list(self._objects.values())

    def list_keys(self):
        """Return a snapshot list of all cache keys."""
        with self._rlock:
            return list(self._objects.keys())

    def get(self, obj):
        """Look up the cached copy of obj. Returns None when absent."""
        key = self._key_func(obj)
        return self.get_by_key(key)

    def get_by_key(self, key):
        """Look up an object by key. Returns None when absent."""
        with self._rlock:
            return self._objects.get(key)



# Copyright 2016 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime

from .dateutil import UTC, TimezoneInfo, format_rfc3339, parse_rfc3339


class DateUtilTest(unittest.TestCase):

    def _parse_rfc3339_test(self, st, y, m, d, h, mn, s, us):
        actual = parse_rfc3339(st)
        expected = datetime(y, m, d, h, mn, s, us, UTC)
        self.assertEqual(expected, actual)

    def test_parse_rfc3339(self):
        self._parse_rfc3339_test("2017-07-25T04:44:21Z",
                                 2017, 7, 25, 4, 44, 21, 0)
        self._parse_rfc3339_test("2017-07-25 04:44:21Z",
                                 2017, 7, 25, 4, 44, 21, 0)
        self._parse_rfc3339_test("2017-07-25T04:44:21",
                                 2017, 7, 25, 4, 44, 21, 0)
        self._parse_rfc3339_test("2017-07-25T04:44:21z",
                                 2017, 7, 25, 4, 44, 21, 0)
        self._parse_rfc3339_test("2017-07-25T04:44:21+03:00",
                                 2017, 7, 25, 1, 44, 21, 0)
        self._parse_rfc3339_test("2017-07-25T04:44:21-03:00",
                                 2017, 7, 25, 7, 44, 21, 0)

        self._parse_rfc3339_test("2017-07-25T04:44:21,005Z",
                                 2017, 7, 25, 4, 44, 21, 5000)
        self._parse_rfc3339_test("2017-07-25T04:44:21.005Z",
                                 2017, 7, 25, 4, 44, 21, 5000)
        self._parse_rfc3339_test("2017-07-25 04:44:21.0050Z",
                                 2017, 7, 25, 4, 44, 21, 5000)
        self._parse_rfc3339_test("2017-07-25T04:44:21.5",
                                 2017, 7, 25, 4, 44, 21, 500000)
        self._parse_rfc3339_test("2017-07-25T04:44:21.005z",
                                 2017, 7, 25, 4, 44, 21, 5000)
        self._parse_rfc3339_test("2017-07-25T04:44:21.005+03:00",
                                 2017, 7, 25, 1, 44, 21, 5000)
        self._parse_rfc3339_test("2017-07-25T04:44:21.005-03:00",
                                 2017, 7, 25, 7, 44, 21, 5000)

    def test_format_rfc3339(self):
        self.assertEqual(
            format_rfc3339(datetime(2017, 7, 25, 4, 44, 21, 0, UTC)),
            "2017-07-25T04:44:21Z")
        self.assertEqual(
            format_rfc3339(datetime(2017, 7, 25, 4, 44, 21, 0,
                                    TimezoneInfo(2, 0))),
            "2017-07-25T02:44:21Z")
        self.assertEqual(
            format_rfc3339(datetime(2017, 7, 25, 4, 44, 21, 0,
                                    TimezoneInfo(-2, 30))),
            "2017-07-25T07:14:21Z")

    def test_parse_rfc3339_invalid_formats(self):
        """Test that invalid RFC3339 formats raise ValueError"""
        invalid_inputs = [
            "2025-13-02T13:37:00Z",        # Invalid month
            "2025-12-32T13:37:00Z",        # Invalid day
            "2025-12-02T25:00:00Z",        # Invalid hour
            "2025-12-02T13:60:00Z",        # Invalid minute
            "2025-12-02T13:37:60Z",        # Invalid second
            "not-a-valid-date",            # Completely invalid
            "",                             # Empty string
            "2025-12-02Z13:37:00",         # Timezone before time
        ]

        for invalid_input in invalid_inputs:
            with self.assertRaises(ValueError):
                parse_rfc3339(invalid_input)



    def test_parse_rfc3339_with_whitespace(self):
        """Test that leading/trailing whitespace is handled"""
        actual = parse_rfc3339("  2017-07-25T04:44:21Z  ")
        expected = datetime(2017, 7, 25, 4, 44, 21, 0, UTC)
        self.assertEqual(expected, actual)

    def test_parse_rfc3339_error_message_clarity(self):
        """Test that error messages are clear and helpful"""
        try:
            parse_rfc3339("invalid-date-format")
        except ValueError as e:
            error_msg = str(e)
            # Verify error message contains helpful information
            self.assertIn("Invalid RFC3339", error_msg)
            self.assertIn("YYYY-MM-DD", error_msg)
            self.assertIn("expected", error_msg)

    def test_parse_rfc3339_handles_none_from_timezone_regex(self):
        """Test parse_rfc3339 handles timezone regex returning None.

        This test addresses the GitHub issue where parse_rfc3339 was
        calling .groups() on None when the timezone regex failed to match,
        causing: 'NoneType' object has no attribute 'groups'

        The fix adds a check to ensure _re_timezone.search() result is
        not None before calling .groups(), and provides a clear error
        message.
        """
        # The main RFC3339 regex allows space in timezone position: [zZ ]
        # If a space ends up in groups[7], it should be handled gracefully
        # Since the current code uses strip(), trailing spaces are removed,
        # but the fix ensures robustness for any edge case

        # Test that space in timezone is treated as UTC (like Z/z)
        actual = parse_rfc3339("2017-07-25 04:44:21")
        expected = datetime(2017, 7, 25, 4, 44, 21, 0, UTC)
        self.assertEqual(expected, actual)



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1IngressSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'default_backend': 'V1IngressBackend',
        'ingress_class_name': 'str',
        'rules': 'list[V1IngressRule]',
        'tls': 'list[V1IngressTLS]'
    }

    attribute_map = {
        'default_backend': 'defaultBackend',
        'ingress_class_name': 'ingressClassName',
        'rules': 'rules',
        'tls': 'tls'
    }

    def __init__(self, default_backend=None, ingress_class_name=None, rules=None, tls=None, local_vars_configuration=None):  # noqa: E501
        """V1IngressSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._default_backend = None
        self._ingress_class_name = None
        self._rules = None
        self._tls = None
        self.discriminator = None

        if default_backend is not None:
            self.default_backend = default_backend
        if ingress_class_name is not None:
            self.ingress_class_name = ingress_class_name
        if rules is not None:
            self.rules = rules
        if tls is not None:
            self.tls = tls

    @property
    def default_backend(self):
        """Gets the default_backend of this V1IngressSpec.  # noqa: E501


        :return: The default_backend of this V1IngressSpec.  # noqa: E501
        :rtype: V1IngressBackend
        """
        return self._default_backend

    @default_backend.setter
    def default_backend(self, default_backend):
        """Sets the default_backend of this V1IngressSpec.


        :param default_backend: The default_backend of this V1IngressSpec.  # noqa: E501
        :type default_backend: V1IngressBackend
        """

        self._default_backend = default_backend

    @property
    def ingress_class_name(self):
        """Gets the ingress_class_name of this V1IngressSpec.  # noqa: E501

        ingressClassName is the name of an IngressClass cluster resource. Ingress controller implementations use this field to know whether they should be serving this Ingress resource, by a transitive connection (controller -> IngressClass -> Ingress resource). Although the `kubernetes.io/ingress.class` annotation (simple constant name) was never formally defined, it was widely supported by Ingress controllers to create a direct binding between Ingress controller and Ingress resources. Newly created Ingress resources should prefer using the field. However, even though the annotation is officially deprecated, for backwards compatibility reasons, ingress controllers should still honor that annotation if present.  # noqa: E501

        :return: The ingress_class_name of this V1IngressSpec.  # noqa: E501
        :rtype: str
        """
        return self._ingress_class_name

    @ingress_class_name.setter
    def ingress_class_name(self, ingress_class_name):
        """Sets the ingress_class_name of this V1IngressSpec.

        ingressClassName is the name of an IngressClass cluster resource. Ingress controller implementations use this field to know whether they should be serving this Ingress resource, by a transitive connection (controller -> IngressClass -> Ingress resource). Although the `kubernetes.io/ingress.class` annotation (simple constant name) was never formally defined, it was widely supported by Ingress controllers to create a direct binding between Ingress controller and Ingress resources. Newly created Ingress resources should prefer using the field. However, even though the annotation is officially deprecated, for backwards compatibility reasons, ingress controllers should still honor that annotation if present.  # noqa: E501

        :param ingress_class_name: The ingress_class_name of this V1IngressSpec.  # noqa: E501
        :type ingress_class_name: str
        """

        self._ingress_class_name = ingress_class_name

    @property
    def rules(self):
        """Gets the rules of this V1IngressSpec.  # noqa: E501

        rules is a list of host rules used to configure the Ingress. If unspecified, or no rule matches, all traffic is sent to the default backend.  # noqa: E501

        :return: The rules of this V1IngressSpec.  # noqa: E501
        :rtype: list[V1IngressRule]
        """
        return self._rules

    @rules.setter
    def rules(self, rules):
        """Sets the rules of this V1IngressSpec.

        rules is a list of host rules used to configure the Ingress. If unspecified, or no rule matches, all traffic is sent to the default backend.  # noqa: E501

        :param rules: The rules of this V1IngressSpec.  # noqa: E501
        :type rules: list[V1IngressRule]
        """

        self._rules = rules

    @property
    def tls(self):
        """Gets the tls of this V1IngressSpec.  # noqa: E501

        tls represents the TLS configuration. Currently the Ingress only supports a single TLS port, 443. If multiple members of this list specify different hosts, they will be multiplexed on the same port according to the hostname specified through the SNI TLS extension, if the ingress controller fulfilling the ingress supports SNI.  # noqa: E501

        :return: The tls of this V1IngressSpec.  # noqa: E501
        :rtype: list[V1IngressTLS]
        """
        return self._tls

    @tls.setter
    def tls(self, tls):
        """Sets the tls of this V1IngressSpec.

        tls represents the TLS configuration. Currently the Ingress only supports a single TLS port, 443. If multiple members of this list specify different hosts, they will be multiplexed on the same port according to the hostname specified through the SNI TLS extension, if the ingress controller fulfilling the ingress supports SNI.  # noqa: E501

        :param tls: The tls of this V1IngressSpec.  # noqa: E501
        :type tls: list[V1IngressTLS]
        """

        self._tls = tls

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1IngressSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1IngressSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha2GangSchedulingPolicy(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'min_count': 'int'
    }

    attribute_map = {
        'min_count': 'minCount'
    }

    def __init__(self, min_count=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha2GangSchedulingPolicy - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._min_count = None
        self.discriminator = None

        self.min_count = min_count

    @property
    def min_count(self):
        """Gets the min_count of this V1alpha2GangSchedulingPolicy.  # noqa: E501

        MinCount is the minimum number of pods that must be schedulable or scheduled at the same time for the scheduler to admit the entire group. It must be a positive integer.  # noqa: E501

        :return: The min_count of this V1alpha2GangSchedulingPolicy.  # noqa: E501
        :rtype: int
        """
        return self._min_count

    @min_count.setter
    def min_count(self, min_count):
        """Sets the min_count of this V1alpha2GangSchedulingPolicy.

        MinCount is the minimum number of pods that must be schedulable or scheduled at the same time for the scheduler to admit the entire group. It must be a positive integer.  # noqa: E501

        :param min_count: The min_count of this V1alpha2GangSchedulingPolicy.  # noqa: E501
        :type min_count: int
        """
        if self.local_vars_configuration.client_side_validation and min_count is None:  # noqa: E501
            raise ValueError("Invalid value for `min_count`, must not be `None`")  # noqa: E501

        self._min_count = min_count

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha2GangSchedulingPolicy):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha2GangSchedulingPolicy):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2026 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Informer implementation for the Kubernetes Python client.

Provides SharedInformer: a background watcher that keeps a local
ObjectCache in sync with the Kubernetes API server and notifies
registered event-handler callbacks.
"""

import logging
import threading
import time

from kubernetes.client.exceptions import ApiException
from kubernetes.watch import Watch

from .cache import ObjectCache, _meta_namespace_key

logger = logging.getLogger(__name__)


# Event types emitted to registered handlers
ADDED = "ADDED"
MODIFIED = "MODIFIED"
DELETED = "DELETED"
BOOKMARK = "BOOKMARK"
ERROR = "ERROR"


class SharedInformer:
    """Watch a Kubernetes resource and maintain a local cache.

    The informer starts a daemon thread that continuously watches the
    given resource via ``list_func``.  On each event the local
    :class:`ObjectCache` is updated and registered
    event-handler callbacks are invoked.

    Parameters
    ----------
    list_func:
        Bound API method used for the initial list **and** as the watch
        source.  It must accept a watch keyword argument (e.g.
        CoreV1Api().list_namespaced_pod).
    namespace:
        Kubernetes namespace to watch.  Pass None for cluster-scoped
        or all-namespace list functions.
    resync_period:
        How often (seconds) to perform a full re-list from the API server.
        Defaults to 0 which disables periodic resyncs.
    label_selector:
        Optional label selector string forwarded to the API server.
    field_selector:
        Optional field selector string forwarded to the API server.
    key_func:
        Optional callable (obj) -> str used to key objects in the
        cache.  Defaults to namespace/name.
    """

    def __init__(
        self,
        list_func,
        namespace=None,
        resync_period=0,
        label_selector=None,
        field_selector=None,
        key_func=None,
    ):
        self._list_func = list_func
        self._namespace = namespace
        self._resync_period = resync_period
        self._label_selector = label_selector
        self._field_selector = field_selector

        self._cache = ObjectCache(key_func=key_func)
        self._handlers = {ADDED: [], MODIFIED: [], DELETED: [], BOOKMARK: [], ERROR: []}
        self._handler_lock = threading.Lock()

        self._watch = None
        self._thread = None
        self._stop_event = threading.Event()
        self._resource_version = None  # most recent RV seen; None forces a full re-list

    # ---------------------------------------------------------------- #
    # Public API                                                        #
    # ---------------------------------------------------------------- #

    @property
    def cache(self):
        """The :class:`ObjectCache` maintained by this informer."""
        return self._cache

    def add_event_handler(self, event_type, handler):
        """Register a callback for a specific event type.

        Parameters
        ----------
        event_type:
            One of :data:`ADDED`, :data:`MODIFIED`, :data:`DELETED`,
            :data:`BOOKMARK` or :data:`ERROR`.
        handler:
            Callable invoked with the event object (or the raw exception for
            ERROR events).
        """
        if event_type not in self._handlers:
            raise ValueError(
                "Unknown event_type {!r}. Use one of: {}".format(
                    event_type, ", ".join(sorted(self._handlers)),
                )
            )
        with self._handler_lock:
            self._handlers[event_type].append(handler)

    def remove_event_handler(self, event_type, handler):
        """Deregister a previously registered *handler*.

        No-op if *handler* is not registered.
        """
        with self._handler_lock:
            try:
                self._handlers[event_type].remove(handler)
            except (KeyError, ValueError):
                pass

    def start(self):
        """Start the background watch loop in a daemon thread.

        Calling :meth:`start` more than once without an intervening
        :meth:`stop` is a no-op.
        """
        if self._thread is not None and self._thread.is_alive():
            return
        self._stop_event.clear()
        self._thread = threading.Thread(
            target=self._run_loop,
            name="SharedInformer",
            daemon=True,
        )
        self._thread.start()

    def stop(self):
        """Ask the background watch loop to stop and join the thread."""
        self._stop_event.set()
        if self._watch is not None:
            self._watch.stop()
        if self._thread is not None:
            self._thread.join()
        self._thread = None

    # ---------------------------------------------------------------- #
    # Internal helpers                                                  #
    # ---------------------------------------------------------------- #

    def _build_kwargs(self):
        kw = {}
        if self._namespace is not None:
            kw["namespace"] = self._namespace
        if self._label_selector is not None:
            kw["label_selector"] = self._label_selector
        if self._field_selector is not None:
            kw["field_selector"] = self._field_selector
        return kw

    def _fire(self, event_type, obj):
        """Execute all registered callbacks for *event_type*, passing *obj*.

        Callbacks are invoked sequentially on the informer's background thread.
        Any exception raised by an individual handler is logged and swallowed so
        that remaining handlers still run.
        """
        with self._handler_lock:
            handlers = list(self._handlers.get(event_type, []))
        for fn in handlers:
            try:
                fn(obj)
            except Exception:
                logger.exception(
                    "Exception in informer handler for %s", event_type
                )

    def _initial_list(self):
        """List all objects and populate the cache, firing ADDED/MODIFIED/DELETED events.

        On the first call (empty cache) every returned item fires ADDED.
        On subsequent calls (resync or after a 410 Gone) the new list is
        diffed against the existing cache:
        * Items absent from the new list fire DELETED.
        * Items present in both fire MODIFIED.
        * Items only in the new list fire ADDED.
        """
        kw = self._build_kwargs()
        resp = self._list_func(**kw)
        items = getattr(resp, "items", []) or []

        # Build key → item map for incoming items.
        new_items_map = {}
        for item in items:
            key = self._cache._key_func(item)
            new_items_map[key] = item

        # Snapshot the old keys before replacing the cache.
        old_keys = set(self._cache.list_keys())

        # Fire DELETED for items no longer present in the new list.
        for key in old_keys:
            if key not in new_items_map:
                old_obj = self._cache.get_by_key(key)
                if old_obj is not None:
                    self._fire(DELETED, old_obj)

        # Atomically replace the cache.
        self._cache._replace_all(items)

        # Fire ADDED for genuinely new items, MODIFIED for existing ones.
        for key, item in new_items_map.items():
            if key in old_keys:
                self._fire(MODIFIED, item)
            else:
                self._fire(ADDED, item)

        rv = None
        meta = getattr(resp, "metadata", None)
        if meta is not None:
            rv = getattr(meta, "resource_version", None)
        self._resource_version = rv or "0"

    def _run_loop(self):
        """Background loop: list then watch, reconnect on errors.

        A full re-list is only performed when ``self._resource_version`` is
        ``None`` (first start or after a 410 Gone response).  On all other
        reconnects the most recent ``resourceVersion`` is reused so that no
        events are missed and the API server does not need to send a full
        object snapshot.
        """
        while not self._stop_event.is_set():
            # Full re-list only when we have no resource version to resume from.
            if self._resource_version is None:
                try:
                    self._initial_list()
                except Exception as exc:
                    logger.exception("Error during initial list; retrying")
                    self._fire(ERROR, exc)
                    self._stop_event.wait(timeout=5)
                    continue

            # Watch loop
            last_resync = time.monotonic()
            self._watch = Watch()
            kw = self._build_kwargs()
            kw["resource_version"] = self._resource_version
            # When a resync period is configured, set a matching server-side
            # watch timeout so that the stream exits after resync_period seconds
            # even if no events arrive.  Without this, a quiet period longer
            # than resync_period would never trigger a resync because the check
            # below only runs when the generator yields an event.
            if self._resync_period > 0:
                kw["timeout_seconds"] = max(1, int(self._resync_period))
            try:
                for event in self._watch.stream(self._list_func, **kw):
                    if self._stop_event.is_set():
                        break
                    evt_type = event.get("type")
                    obj = event.get("object")
                    # Sync the most recent resource version from the Watch
                    # instance (updated by unmarshal_event before yielding).
                    # Do this before firing handlers so consumers that wake on
                    # an event immediately see the advanced resource version.
                    if self._watch is not None and self._watch.resource_version:
                        self._resource_version = self._watch.resource_version
                    if evt_type == ADDED:
                        self._cache._put(obj)
                        self._fire(ADDED, obj)
                    elif evt_type == MODIFIED:
                        self._cache._put(obj)
                        self._fire(MODIFIED, obj)
                    elif evt_type == DELETED:
                        self._cache._remove(obj)
                        self._fire(DELETED, obj)
                    elif evt_type == BOOKMARK:
                        # BOOKMARK events carry an updated resource version but
                        # no object state change; the Watch instance already
                        # records the new resource_version internally.
                        self._fire(BOOKMARK, event.get("raw_object", obj))
                    elif evt_type == ERROR:
                        self._fire(ERROR, obj)
            except ApiException as exc:
                if exc.status == 410:
                    # The stored resource version is too old; force a full re-list.
                    logger.warning(
                        "Watch expired (410 Gone); will re-list from scratch"
                    )
                    self._resource_version = None
                else:
                    logger.warning(
                        "Watch stream ended with ApiException (status=%s); reconnecting",
                        exc.status,
                    )
                self._fire(ERROR, exc)
            except Exception as exc:
                logger.exception("Unexpected error in watch loop; reconnecting")
                self._fire(ERROR, exc)
            finally:
                # Capture the most recent resource version seen by the Watch
                # (updated on every ADDED/MODIFIED/DELETED/BOOKMARK event) so
                # that the next watch connection can resume without re-listing.
                # Do not overwrite a None that was set by a 410 handler above.
                if (
                    self._resource_version is not None
                    and self._watch is not None
                    and self._watch.resource_version
                ):
                    self._resource_version = self._watch.resource_version
                self._watch = None

            # Periodic resync: after the watch stream exits (whether due to the
            # server-side timeout_seconds, a stop request, or an error) check if
            # a resync is due.  This path is what actually fires the resync when
            # the cluster is quiet and no events arrive for resync_period seconds.
            if (
                not self._stop_event.is_set()
                and self._resource_version is not None  # 410 already schedules a re-list
                and self._resync_period > 0
                and (time.monotonic() - last_resync) >= self._resync_period
            ):
                logger.debug("Informer resync triggered")
                try:
                    self._initial_list()
                except Exception as exc:
                    logger.exception("Error during resync list; continuing")
                    self._fire(ERROR, exc)



# Copyright 2016 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import unittest
from unittest.mock import Mock, call

from kubernetes import client, config
from kubernetes.client import ApiException

from .watch import Watch


class WatchTests(unittest.TestCase):
    def setUp(self):
        # counter for a test that needs test global state
        self.callcount = 0

    def test_watch_with_decode(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                '{"type": "ADDED", "object": {"metadata": {"name": "test1",'
                '"resourceVersion": "1"}, "spec": {}, "status": {}}}\n',
                '{"type": "ADDED", "object": {"metadata": {"name": "test2",'
                '"resourceVersion": "2"}, "spec": {}, "sta',
                'tus": {}}}\n'
                '{"type": "ADDED", "object": {"metadata": {"name": "test3",'
                '"resourceVersion": "3"}, "spec": {}, "status": {}}}\n',
                'should_not_happened\n'])

        fake_api = Mock()
        fake_api.get_namespaces = Mock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'

        w = Watch()
        count = 1
        for e in w.stream(fake_api.get_namespaces):
            self.assertEqual("ADDED", e['type'])
            # make sure decoder worked and we got a model with the right name
            self.assertEqual("test%d" % count, e['object'].metadata.name)
            # make sure decoder worked and updated Watch.resource_version
            self.assertEqual(
                "%d" % count, e['object'].metadata.resource_version)
            self.assertEqual("%d" % count, w.resource_version)
            count += 1
            # make sure we can stop the watch and the last event with won't be
            # returned
            if count == 4:
                w.stop()

        # make sure that all three records were consumed by the stream
        self.assertEqual(4, count)

        fake_api.get_namespaces.assert_called_once_with(
            _preload_content=False, watch=True)
        fake_resp.stream.assert_called_once_with(
            amt=None, decode_content=False)
        fake_resp.close.assert_called_once()
        fake_resp.release_conn.assert_called_once()

    def test_watch_with_interspersed_newlines(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                '\n',
                '{"type": "ADDED", "object": {"metadata":',
                '{"name": "test1","resourceVersion": "1"}}}\n{"type": "ADDED", ',
                '"object": {"metadata": {"name": "test2", "resourceVersion": "2"}}}\n',
                '\n',
                '',
                '{"type": "ADDED", "object": {"metadata": {"name": "test3", "resourceVersion": "3"}}}\n',
                '\n\n\n',
                '\n',
            ])

        fake_api = Mock()
        fake_api.get_namespaces = Mock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'

        w = Watch()
        count = 0

        # Consume all test events from the mock service, stopping when no more data is available.
        # Note that "timeout_seconds" below is not a timeout; rather, it disables retries and is
        # the only way to do so. Without that, the stream will re-read the test data forever.
        for e in w.stream(fake_api.get_namespaces, timeout_seconds=1):
            # Here added a statement for exception for empty lines.
            if e is None:
                continue
            count += 1
            self.assertEqual("test%d" % count, e['object'].metadata.name)
        self.assertEqual(3, count)

    def test_watch_with_multibyte_utf8(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                # two-byte utf-8 character
                '{"type":"MODIFIED","object":{"data":{"utf-8":"© 1"},"metadata":{"name":"test1","resourceVersion":"1"}}}\n',
                # same copyright character expressed as bytes
                b'{"type":"MODIFIED","object":{"data":{"utf-8":"\xC2\xA9 2"},"metadata":{"name":"test2","resourceVersion":"2"}}}\n'
                # same copyright character with bytes split across two stream chunks
                b'{"type":"MODIFIED","object":{"data":{"utf-8":"\xC2',
                b'\xA9 3"},"metadata":{"n',
                # more chunks of the same event, sent as a mix of bytes and strings
                'ame":"test3","resourceVersion":"3"',
                '}}}',
                b'\n'
            ])

        fake_api = Mock()
        fake_api.get_configmaps = Mock(return_value=fake_resp)
        fake_api.get_configmaps.__doc__ = ':rtype: V1ConfigMapList'

        w = Watch()
        count = 0

        # Consume all test events from the mock service, stopping when no more data is available.
        # Note that "timeout_seconds" below is not a timeout; rather, it disables retries and is
        # the only way to do so. Without that, the stream will re-read the test data forever.
        for event in w.stream(fake_api.get_configmaps, timeout_seconds=1):
            count += 1
            self.assertEqual("MODIFIED", event['type'])
            self.assertEqual("test%d" % count, event['object'].metadata.name)
            self.assertEqual("© %d" % count, event['object'].data["utf-8"])
            self.assertEqual(
                "%d" % count, event['object'].metadata.resource_version)
            self.assertEqual("%d" % count, w.resource_version)
        self.assertEqual(3, count)

    def test_watch_with_invalid_utf8(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            # test 1 uses 1 invalid utf-8 byte
            # test 2 uses a sequence of 2 invalid utf-8 bytes
            # test 3 uses a sequence of 3 invalid utf-8 bytes
            return_value=[
                # utf-8 sequence for 😄 is \xF0\x9F\x98\x84
                # all other sequences below are invalid
                # ref: https://www.w3.org/2001/06/utf-8-wrong/UTF-8-test.html
                b'{"type":"MODIFIED","object":{"data":{"utf-8":"\xF0\x9F\x98\x84 1","invalid":"\x80 1"},"metadata":{"name":"test1"}}}\n',
                b'{"type":"MODIFIED","object":{"data":{"utf-8":"\xF0\x9F\x98\x84 2","invalid":"\xC0\xAF 2"},"metadata":{"name":"test2"}}}\n',
                # mix bytes/strings and split byte sequences across chunks
                b'{"type":"MODIFIED","object":{"data":{"utf-8":"\xF0\x9F\x98',
                b'\x84 ',
                b'',
                b'3","invalid":"\xE0\x80',
                b'\xAF ',
                '3"},"metadata":{"n',
                'ame":"test3"',
                '}}}',
                b'\n'
            ])

        fake_api = Mock()
        fake_api.get_configmaps = Mock(return_value=fake_resp)
        fake_api.get_configmaps.__doc__ = ':rtype: V1ConfigMapList'

        w = Watch()
        count = 0

        # Consume all test events from the mock service, stopping when no more data is available.
        # Note that "timeout_seconds" below is not a timeout; rather, it disables retries and is
        # the only way to do so. Without that, the stream will re-read the test data forever.
        for event in w.stream(fake_api.get_configmaps, timeout_seconds=1):
            count += 1
            self.assertEqual("MODIFIED", event['type'])
            self.assertEqual("test%d" % count, event['object'].metadata.name)
            self.assertEqual("😄 %d" % count, event['object'].data["utf-8"])
            # expect N replacement characters in test N
            self.assertEqual("� %d".replace('�', '�'*count) %
                             count, event['object'].data["invalid"])
        self.assertEqual(3, count)

    def test_watch_for_follow(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                'log_line_1\n',
                'log_line_2\n'])

        fake_api = Mock()
        fake_api.read_namespaced_pod_log = Mock(return_value=fake_resp)
        fake_api.read_namespaced_pod_log.__doc__ = ':param bool follow:\n:rtype: str'

        w = Watch()
        count = 1
        for e in w.stream(fake_api.read_namespaced_pod_log):
            self.assertEqual("log_line_1", e)
            count += 1
            # make sure we can stop the watch and the last event with won't be
            # returned
            if count == 2:
                w.stop()

        fake_api.read_namespaced_pod_log.assert_called_once_with(
            _preload_content=False, follow=True)
        fake_resp.stream.assert_called_once_with(
            amt=None, decode_content=False)
        fake_resp.close.assert_called_once()
        fake_resp.release_conn.assert_called_once()

    def test_watch_resource_version_set(self):
        # https://github.com/kubernetes-client/python/issues/700
        # ensure watching from a resource version does reset to resource
        # version 0 after k8s resets the watch connection
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        values = [
            '{"type": "ADDED", "object": {"metadata": {"name": "test1",'
            '"resourceVersion": "1"}, "spec": {}, "status": {}}}\n',
            '{"type": "ADDED", "object": {"metadata": {"name": "test2",'
            '"resourceVersion": "2"}, "spec": {}, "sta',
            'tus": {}}}\n'
            '{"type": "ADDED", "object": {"metadata": {"name": "test3",'
            '"resourceVersion": "3"}, "spec": {}, "status": {}}}\n'
        ]

        # return nothing on the first call and values on the second
        # this emulates a watch from a rv that returns nothing in the first k8s
        # watch reset and values later

        def get_values(*args, **kwargs):
            self.callcount += 1
            if self.callcount == 1:
                return []
            else:
                return values

        fake_resp.stream = Mock(
            side_effect=get_values)

        fake_api = Mock()
        fake_api.get_namespaces = Mock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'

        w = Watch()
        # ensure we keep our requested resource version or the version latest
        # returned version when the existing versions are older than the
        # requested version
        # needed for the list existing objects, then watch from there use case
        calls = []

        iterations = 2
        # first two calls must use the passed rv, the first call is a
        # "reset" and does not actually return anything
        # the second call must use the same rv but will return values
        # (with a wrong rv but a real cluster would behave correctly)
        # calls following that will use the rv from those returned values
        calls.append(call(_preload_content=False, watch=True,
                          resource_version="5"))
        calls.append(call(_preload_content=False, watch=True,
                          resource_version="5"))
        for i in range(iterations):
            # ideally we want 5 here but as rv must be treated as an
            # opaque value we cannot interpret it and order it so rely
            # on k8s returning the events completely and in order
            calls.append(call(_preload_content=False, watch=True,
                              resource_version="3"))

        for c, e in enumerate(w.stream(fake_api.get_namespaces,
                                       resource_version="5")):
            if c == len(values) * iterations:
                w.stop()

        # check calls are in the list, gives good error output
        fake_api.get_namespaces.assert_has_calls(calls)
        # more strict test with worse error message
        self.assertEqual(fake_api.get_namespaces.mock_calls, calls)

    def test_watch_stream_twice(self):
        w = Watch(float)
        for step in ['first', 'second']:
            fake_resp = Mock()
            fake_resp.close = Mock()
            fake_resp.release_conn = Mock()
            fake_resp.stream = Mock(
                return_value=['{"type": "ADDED", "object": 1}\n'] * 4)

            fake_api = Mock()
            fake_api.get_namespaces = Mock(return_value=fake_resp)
            fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'

            count = 1
            for e in w.stream(fake_api.get_namespaces):
                count += 1
                if count == 3:
                    w.stop()

            self.assertEqual(count, 3)
            fake_api.get_namespaces.assert_called_once_with(
                _preload_content=False, watch=True)
            fake_resp.stream.assert_called_once_with(
                amt=None, decode_content=False)
            fake_resp.close.assert_called_once()
            fake_resp.release_conn.assert_called_once()

    def test_watch_stream_loop(self):
        w = Watch(float)

        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=['{"type": "ADDED", "object": 1}\n'])

        fake_api = Mock()
        fake_api.get_namespaces = Mock(return_value=fake_resp)
        fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'

        count = 0

        # when timeout_seconds is set, auto-exist when timeout reaches
        for e in w.stream(fake_api.get_namespaces, timeout_seconds=1):
            count = count + 1
        self.assertEqual(count, 1)

        # when no timeout_seconds, only exist when w.stop() is called
        for e in w.stream(fake_api.get_namespaces):
            count = count + 1
            if count == 2:
                w.stop()

        self.assertEqual(count, 2)
        self.assertEqual(fake_api.get_namespaces.call_count, 2)
        self.assertEqual(fake_resp.stream.call_count, 2)
        self.assertEqual(fake_resp.close.call_count, 2)
        self.assertEqual(fake_resp.release_conn.call_count, 2)

    def test_unmarshal_with_float_object(self):
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": 1}', 'float')
        self.assertEqual("ADDED", event['type'])
        self.assertEqual(1.0, event['object'])
        self.assertTrue(isinstance(event['object'], float))
        self.assertEqual(1, event['raw_object'])

    def test_unmarshal_with_no_return_type(self):
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": ["test1"]}',
                                  None)
        self.assertEqual("ADDED", event['type'])
        self.assertEqual(["test1"], event['object'])
        self.assertEqual(["test1"], event['raw_object'])

    def test_unmarshal_with_custom_object(self):
        w = Watch()
        event = w.unmarshal_event('{"type": "ADDED", "object": {"apiVersion":'
                                  '"test.com/v1beta1","kind":"foo","metadata":'
                                  '{"name": "bar", "resourceVersion": "1"}}}',
                                  'object')
        self.assertEqual("ADDED", event['type'])
        # make sure decoder deserialized json into dictionary and updated
        # Watch.resource_version
        self.assertTrue(isinstance(event['object'], dict))
        self.assertEqual("1", event['object']['metadata']['resourceVersion'])
        self.assertEqual("1", w.resource_version)

    def test_unmarshal_with_bookmark(self):
        w = Watch()
        event = w.unmarshal_event(
            '{"type":"BOOKMARK","object":{"kind":"Job","apiVersion":"batch/v1"'
            ',"metadata":{"resourceVersion":"1"},"spec":{"template":{'
            '"metadata":{},"spec":{"containers":null}}},"status":{}}}',
            'V1Job')
        self.assertEqual("BOOKMARK", event['type'])
        self.assertEqual("1", w.resource_version)

    def test_unmarshal_with_bookmark_metadata_not_in_dict(self):
        w = Watch()
        event = w.unmarshal_event(
            '{"type":"BOOKMARK","object":{"metadata": "not-a-dict"}}',
            'V1Job')
        self.assertEqual("BOOKMARK", event['type'])
        self.assertEqual(None, w.resource_version)

    def test_unmarshal_with_bookmark_metadata_without_resource_version(self):
        w = Watch()
        event = w.unmarshal_event(
            '{"type":"BOOKMARK","object":{"metadata": {"name": "foo"}}}',
            'V1Job')
        self.assertEqual("BOOKMARK", event['type'])
        self.assertEqual(None, w.resource_version)

    def test_watch_with_exception(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(side_effect=KeyError('expected'))

        fake_api = Mock()
        fake_api.get_thing = Mock(return_value=fake_resp)

        w = Watch()
        try:
            for _ in w.stream(fake_api.get_thing):
                self.fail(self, "Should fail on exception.")
        except KeyError:
            pass
            # expected

        fake_api.get_thing.assert_called_once_with(
            _preload_content=False, watch=True)
        fake_resp.stream.assert_called_once_with(
            amt=None, decode_content=False)
        fake_resp.close.assert_called_once()
        fake_resp.release_conn.assert_called_once()

    def test_watch_with_error_event(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                '{"type": "ERROR", "object": {"code": 410, '
                '"reason": "Gone", "message": "error message"}}\n'])

        fake_api = Mock()
        fake_api.get_thing = Mock(return_value=fake_resp)

        w = Watch()
        # No events are generated when no initial resourceVersion is passed
        # No retry is attempted either, preventing an ApiException
        assert not list(w.stream(fake_api.get_thing))

        fake_api.get_thing.assert_called_once_with(
            _preload_content=False, watch=True)
        fake_resp.stream.assert_called_once_with(
            amt=None, decode_content=False)
        fake_resp.close.assert_called_once()
        fake_resp.release_conn.assert_called_once()

    def test_watch_retries_on_error_event(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                '{"type": "ERROR", "object": {"code": 410, '
                '"reason": "Gone", "message": "error message"}}\n'])

        fake_api = Mock()
        fake_api.get_thing = Mock(return_value=fake_resp)

        w = Watch()
        try:
            for _ in w.stream(fake_api.get_thing, resource_version=0):
                self.fail(self, "Should fail with ApiException.")
        except client.rest.ApiException:
            pass

        # Two calls should be expected during a retry
        fake_api.get_thing.assert_has_calls(
            [call(resource_version=0, _preload_content=False, watch=True)] * 2)
        fake_resp.stream.assert_has_calls(
            [call(amt=None, decode_content=False)] * 2)
        assert fake_resp.close.call_count == 2
        assert fake_resp.release_conn.call_count == 2

    def test_watch_with_error_event_and_timeout_param(self):
        fake_resp = Mock()
        fake_resp.close = Mock()
        fake_resp.release_conn = Mock()
        fake_resp.stream = Mock(
            return_value=[
                '{"type": "ERROR", "object": {"code": 410, '
                '"reason": "Gone", "message": "error message"}}\n'])

        fake_api = Mock()
        fake_api.get_thing = Mock(return_value=fake_resp)

        w = Watch()
        try:
            for _ in w.stream(fake_api.get_thing, timeout_seconds=10):
                self.fail(self, "Should fail with ApiException.")
        except client.rest.ApiException:
            pass

        fake_api.get_thing.assert_called_once_with(
            _preload_content=False, watch=True, timeout_seconds=10)
        fake_resp.stream.assert_called_once_with(
            amt=None, decode_content=False)
        fake_resp.close.assert_called_once()
        fake_resp.release_conn.assert_called_once()
    
    @classmethod
    def setUpClass(cls):
        cls.api = Mock()
        cls.namespace = "default"

    def test_pod_log_empty_lines(self):
        pod_name = "demo-bug"
        
        try:
            self.api.create_namespaced_pod = Mock()
            self.api.read_namespaced_pod = Mock()
            self.api.delete_namespaced_pod = Mock()
            self.api.read_namespaced_pod_log = Mock()

            #pod creating step
            self.api.create_namespaced_pod.return_value = None
            
            #Checking pod status
            mock_pod = Mock()
            mock_pod.status.phase = "Running"
            self.api.read_namespaced_pod.return_value = mock_pod
            
            # Printing at pod output
            self.api.read_namespaced_pod_log.return_value = iter(["Hello from Docker\n"])

            # Wait for the pod to reach 'Running'
            timeout = 60
            start_time = time.time()
            while time.time() - start_time < timeout:
                pod = self.api.read_namespaced_pod(name=pod_name, namespace=self.namespace)
                if pod.status.phase == "Running":
                    break
                time.sleep(2)
            else:
                self.fail("Pod did not reach 'Running' state within timeout")

            # Reading and streaming logs using Watch (mocked)
            w = Watch()
            log_output = []
            #Mock logs used for this test
            w.stream = Mock(return_value=[
                        "Hello from Docker",
                        "",
                        "",
                        "\n\n",
                        "Another log line",
                        "",
                        "\n",
                        "Final log"
                    ])
            for event in w.stream(self.api.read_namespaced_pod_log, name=pod_name, namespace=self.namespace, follow=True):
                log_output.append(event)
                print(event)

            # Print outputs
            print(f"Captured logs: {log_output}") 
            # self.assertTrue(any("Hello from Docker" in line for line in log_output))
            # self.assertTrue(any(line.strip() == "" for line in log_output), "No empty lines found in logs")
            expected_log = [
                "Hello from Docker",
                "",
                "",
                "\n\n",
                "Another log line",
                "",
                "\n",
                "Final log"
            ]
            
            self.assertEqual(log_output, expected_log, "Captured logs do not match expected logs")

        except ApiException as e:
            self.fail(f"Kubernetes API exception: {e}")
        finally:
            #checking pod is calling for delete
            self.api.delete_namespaced_pod(name=pod_name, namespace=self.namespace)
            self.api.delete_namespaced_pod.assert_called_once_with(name=pod_name, namespace=self.namespace)

#    Comment out the test below, it does not work currently.
#    def test_watch_with_deserialize_param(self):
#        """test watch.stream() deserialize param"""
#        # prepare test data
#        test_json = '{"type": "ADDED", "object": {"metadata": {"name": "test1", "resourceVersion": "1"}, "spec": {}, "status": {}}}'
#        fake_resp = Mock()
#        fake_resp.close = Mock()
#        fake_resp.release_conn = Mock()
#        fake_resp.stream = Mock(return_value=[test_json + '\n'])
#    
#        fake_api = Mock()
#        fake_api.get_namespaces = Mock(return_value=fake_resp)
#        fake_api.get_namespaces.__doc__ = ':rtype: V1NamespaceList'
#    
#        # test case with deserialize=True
#        w = Watch()
#        for e in w.stream(fake_api.get_namespaces, deserialize=True):
#            self.assertEqual("ADDED", e['type'])
#            # Verify that the object is deserialized correctly
#            self.assertTrue(hasattr(e['object'], 'metadata'))
#            self.assertEqual("test1", e['object'].metadata.name)
#            self.assertEqual("1", e['object'].metadata.resource_version)
#            # Verify that the original object is saved
#            self.assertEqual(json.loads(test_json)['object'], e['raw_object'])
#    
#        # test case with deserialize=False
#        w = Watch()
#        for e in w.stream(fake_api.get_namespaces, deserialize=False):
#            self.assertEqual("ADDED", e['type'])
#            # The validation object remains in the original dictionary format
#            self.assertIsInstance(e['object'], dict)
#            self.assertEqual("test1", e['object']['metadata']['name'])
#            self.assertEqual("1", e['object']['metadata']['resourceVersion'])
#    
#        # verify the api is called twice
#        fake_api.get_namespaces.assert_has_calls([
#            call(_preload_content=False, watch=True),
#            call(_preload_content=False, watch=True)
#        ])
    

if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha2PodGroupTemplate(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'disruption_mode': 'str',
        'name': 'str',
        'priority': 'int',
        'priority_class_name': 'str',
        'resource_claims': 'list[V1alpha2PodGroupResourceClaim]',
        'scheduling_constraints': 'V1alpha2PodGroupSchedulingConstraints',
        'scheduling_policy': 'V1alpha2PodGroupSchedulingPolicy'
    }

    attribute_map = {
        'disruption_mode': 'disruptionMode',
        'name': 'name',
        'priority': 'priority',
        'priority_class_name': 'priorityClassName',
        'resource_claims': 'resourceClaims',
        'scheduling_constraints': 'schedulingConstraints',
        'scheduling_policy': 'schedulingPolicy'
    }

    def __init__(self, disruption_mode=None, name=None, priority=None, priority_class_name=None, resource_claims=None, scheduling_constraints=None, scheduling_policy=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha2PodGroupTemplate - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._disruption_mode = None
        self._name = None
        self._priority = None
        self._priority_class_name = None
        self._resource_claims = None
        self._scheduling_constraints = None
        self._scheduling_policy = None
        self.discriminator = None

        if disruption_mode is not None:
            self.disruption_mode = disruption_mode
        self.name = name
        if priority is not None:
            self.priority = priority
        if priority_class_name is not None:
            self.priority_class_name = priority_class_name
        if resource_claims is not None:
            self.resource_claims = resource_claims
        if scheduling_constraints is not None:
            self.scheduling_constraints = scheduling_constraints
        self.scheduling_policy = scheduling_policy

    @property
    def disruption_mode(self):
        """Gets the disruption_mode of this V1alpha2PodGroupTemplate.  # noqa: E501

        DisruptionMode defines the mode in which a given PodGroup can be disrupted. One of Pod, PodGroup. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :return: The disruption_mode of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: str
        """
        return self._disruption_mode

    @disruption_mode.setter
    def disruption_mode(self, disruption_mode):
        """Sets the disruption_mode of this V1alpha2PodGroupTemplate.

        DisruptionMode defines the mode in which a given PodGroup can be disrupted. One of Pod, PodGroup. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :param disruption_mode: The disruption_mode of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type disruption_mode: str
        """

        self._disruption_mode = disruption_mode

    @property
    def name(self):
        """Gets the name of this V1alpha2PodGroupTemplate.  # noqa: E501

        Name is a unique identifier for the PodGroupTemplate within the Workload. It must be a DNS label. This field is immutable.  # noqa: E501

        :return: The name of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this V1alpha2PodGroupTemplate.

        Name is a unique identifier for the PodGroupTemplate within the Workload. It must be a DNS label. This field is immutable.  # noqa: E501

        :param name: The name of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def priority(self):
        """Gets the priority of this V1alpha2PodGroupTemplate.  # noqa: E501

        Priority is the value of priority of pod groups created from this template. Various system components use this field to find the priority of the pod group. When Priority Admission Controller is enabled, it prevents users from setting this field. The admission controller populates this field from PriorityClassName. The higher the value, the higher the priority. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :return: The priority of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: int
        """
        return self._priority

    @priority.setter
    def priority(self, priority):
        """Sets the priority of this V1alpha2PodGroupTemplate.

        Priority is the value of priority of pod groups created from this template. Various system components use this field to find the priority of the pod group. When Priority Admission Controller is enabled, it prevents users from setting this field. The admission controller populates this field from PriorityClassName. The higher the value, the higher the priority. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :param priority: The priority of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type priority: int
        """

        self._priority = priority

    @property
    def priority_class_name(self):
        """Gets the priority_class_name of this V1alpha2PodGroupTemplate.  # noqa: E501

        PriorityClassName indicates the priority that should be considered when scheduling a pod group created from this template. If no priority class is specified, admission control can set this to the global default priority class if it exists. Otherwise, pod groups created from this template will have the priority set to zero. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :return: The priority_class_name of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: str
        """
        return self._priority_class_name

    @priority_class_name.setter
    def priority_class_name(self, priority_class_name):
        """Sets the priority_class_name of this V1alpha2PodGroupTemplate.

        PriorityClassName indicates the priority that should be considered when scheduling a pod group created from this template. If no priority class is specified, admission control can set this to the global default priority class if it exists. Otherwise, pod groups created from this template will have the priority set to zero. This field is available only when the WorkloadAwarePreemption feature gate is enabled.  # noqa: E501

        :param priority_class_name: The priority_class_name of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type priority_class_name: str
        """

        self._priority_class_name = priority_class_name

    @property
    def resource_claims(self):
        """Gets the resource_claims of this V1alpha2PodGroupTemplate.  # noqa: E501

        ResourceClaims defines which ResourceClaims may be shared among Pods in the group. Pods consume the devices allocated to a PodGroup's claim by defining a claim in its own Spec.ResourceClaims that matches the PodGroup's claim exactly. The claim must have the same name and refer to the same ResourceClaim or ResourceClaimTemplate.  This is an alpha-level field and requires that the DRAWorkloadResourceClaims feature gate is enabled.  This field is immutable.  # noqa: E501

        :return: The resource_claims of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: list[V1alpha2PodGroupResourceClaim]
        """
        return self._resource_claims

    @resource_claims.setter
    def resource_claims(self, resource_claims):
        """Sets the resource_claims of this V1alpha2PodGroupTemplate.

        ResourceClaims defines which ResourceClaims may be shared among Pods in the group. Pods consume the devices allocated to a PodGroup's claim by defining a claim in its own Spec.ResourceClaims that matches the PodGroup's claim exactly. The claim must have the same name and refer to the same ResourceClaim or ResourceClaimTemplate.  This is an alpha-level field and requires that the DRAWorkloadResourceClaims feature gate is enabled.  This field is immutable.  # noqa: E501

        :param resource_claims: The resource_claims of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type resource_claims: list[V1alpha2PodGroupResourceClaim]
        """

        self._resource_claims = resource_claims

    @property
    def scheduling_constraints(self):
        """Gets the scheduling_constraints of this V1alpha2PodGroupTemplate.  # noqa: E501


        :return: The scheduling_constraints of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: V1alpha2PodGroupSchedulingConstraints
        """
        return self._scheduling_constraints

    @scheduling_constraints.setter
    def scheduling_constraints(self, scheduling_constraints):
        """Sets the scheduling_constraints of this V1alpha2PodGroupTemplate.


        :param scheduling_constraints: The scheduling_constraints of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type scheduling_constraints: V1alpha2PodGroupSchedulingConstraints
        """

        self._scheduling_constraints = scheduling_constraints

    @property
    def scheduling_policy(self):
        """Gets the scheduling_policy of this V1alpha2PodGroupTemplate.  # noqa: E501


        :return: The scheduling_policy of this V1alpha2PodGroupTemplate.  # noqa: E501
        :rtype: V1alpha2PodGroupSchedulingPolicy
        """
        return self._scheduling_policy

    @scheduling_policy.setter
    def scheduling_policy(self, scheduling_policy):
        """Sets the scheduling_policy of this V1alpha2PodGroupTemplate.


        :param scheduling_policy: The scheduling_policy of this V1alpha2PodGroupTemplate.  # noqa: E501
        :type scheduling_policy: V1alpha2PodGroupSchedulingPolicy
        """
        if self.local_vars_configuration.client_side_validation and scheduling_policy is None:  # noqa: E501
            raise ValueError("Invalid value for `scheduling_policy`, must not be `None`")  # noqa: E501

        self._scheduling_policy = scheduling_policy

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha2PodGroupTemplate):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha2PodGroupTemplate):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1ResourceSliceSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'all_nodes': 'bool',
        'devices': 'list[V1Device]',
        'driver': 'str',
        'node_name': 'str',
        'node_selector': 'V1NodeSelector',
        'per_device_node_selection': 'bool',
        'pool': 'V1ResourcePool',
        'shared_counters': 'list[V1CounterSet]'
    }

    attribute_map = {
        'all_nodes': 'allNodes',
        'devices': 'devices',
        'driver': 'driver',
        'node_name': 'nodeName',
        'node_selector': 'nodeSelector',
        'per_device_node_selection': 'perDeviceNodeSelection',
        'pool': 'pool',
        'shared_counters': 'sharedCounters'
    }

    def __init__(self, all_nodes=None, devices=None, driver=None, node_name=None, node_selector=None, per_device_node_selection=None, pool=None, shared_counters=None, local_vars_configuration=None):  # noqa: E501
        """V1ResourceSliceSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._all_nodes = None
        self._devices = None
        self._driver = None
        self._node_name = None
        self._node_selector = None
        self._per_device_node_selection = None
        self._pool = None
        self._shared_counters = None
        self.discriminator = None

        if all_nodes is not None:
            self.all_nodes = all_nodes
        if devices is not None:
            self.devices = devices
        self.driver = driver
        if node_name is not None:
            self.node_name = node_name
        if node_selector is not None:
            self.node_selector = node_selector
        if per_device_node_selection is not None:
            self.per_device_node_selection = per_device_node_selection
        self.pool = pool
        if shared_counters is not None:
            self.shared_counters = shared_counters

    @property
    def all_nodes(self):
        """Gets the all_nodes of this V1ResourceSliceSpec.  # noqa: E501

        AllNodes indicates that all nodes have access to the resources in the pool.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set.  # noqa: E501

        :return: The all_nodes of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: bool
        """
        return self._all_nodes

    @all_nodes.setter
    def all_nodes(self, all_nodes):
        """Sets the all_nodes of this V1ResourceSliceSpec.

        AllNodes indicates that all nodes have access to the resources in the pool.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set.  # noqa: E501

        :param all_nodes: The all_nodes of this V1ResourceSliceSpec.  # noqa: E501
        :type all_nodes: bool
        """

        self._all_nodes = all_nodes

    @property
    def devices(self):
        """Gets the devices of this V1ResourceSliceSpec.  # noqa: E501

        Devices lists some or all of the devices in this pool.  Must not have more than 128 entries. If any device uses taints or consumes counters the limit is 64.  Only one of Devices and SharedCounters can be set in a ResourceSlice.  # noqa: E501

        :return: The devices of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: list[V1Device]
        """
        return self._devices

    @devices.setter
    def devices(self, devices):
        """Sets the devices of this V1ResourceSliceSpec.

        Devices lists some or all of the devices in this pool.  Must not have more than 128 entries. If any device uses taints or consumes counters the limit is 64.  Only one of Devices and SharedCounters can be set in a ResourceSlice.  # noqa: E501

        :param devices: The devices of this V1ResourceSliceSpec.  # noqa: E501
        :type devices: list[V1Device]
        """

        self._devices = devices

    @property
    def driver(self):
        """Gets the driver of this V1ResourceSliceSpec.  # noqa: E501

        Driver identifies the DRA driver providing the capacity information. A field selector can be used to list only ResourceSlice objects with a certain driver name.  Must be a DNS subdomain and should end with a DNS domain owned by the vendor of the driver. It should use only lower case characters. This field is immutable.  # noqa: E501

        :return: The driver of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: str
        """
        return self._driver

    @driver.setter
    def driver(self, driver):
        """Sets the driver of this V1ResourceSliceSpec.

        Driver identifies the DRA driver providing the capacity information. A field selector can be used to list only ResourceSlice objects with a certain driver name.  Must be a DNS subdomain and should end with a DNS domain owned by the vendor of the driver. It should use only lower case characters. This field is immutable.  # noqa: E501

        :param driver: The driver of this V1ResourceSliceSpec.  # noqa: E501
        :type driver: str
        """
        if self.local_vars_configuration.client_side_validation and driver is None:  # noqa: E501
            raise ValueError("Invalid value for `driver`, must not be `None`")  # noqa: E501

        self._driver = driver

    @property
    def node_name(self):
        """Gets the node_name of this V1ResourceSliceSpec.  # noqa: E501

        NodeName identifies the node which provides the resources in this pool. A field selector can be used to list only ResourceSlice objects belonging to a certain node.  This field can be used to limit access from nodes to ResourceSlices with the same node name. It also indicates to autoscalers that adding new nodes of the same type as some old node might also make new resources available.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set. This field is immutable.  # noqa: E501

        :return: The node_name of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: str
        """
        return self._node_name

    @node_name.setter
    def node_name(self, node_name):
        """Sets the node_name of this V1ResourceSliceSpec.

        NodeName identifies the node which provides the resources in this pool. A field selector can be used to list only ResourceSlice objects belonging to a certain node.  This field can be used to limit access from nodes to ResourceSlices with the same node name. It also indicates to autoscalers that adding new nodes of the same type as some old node might also make new resources available.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set. This field is immutable.  # noqa: E501

        :param node_name: The node_name of this V1ResourceSliceSpec.  # noqa: E501
        :type node_name: str
        """

        self._node_name = node_name

    @property
    def node_selector(self):
        """Gets the node_selector of this V1ResourceSliceSpec.  # noqa: E501


        :return: The node_selector of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: V1NodeSelector
        """
        return self._node_selector

    @node_selector.setter
    def node_selector(self, node_selector):
        """Sets the node_selector of this V1ResourceSliceSpec.


        :param node_selector: The node_selector of this V1ResourceSliceSpec.  # noqa: E501
        :type node_selector: V1NodeSelector
        """

        self._node_selector = node_selector

    @property
    def per_device_node_selection(self):
        """Gets the per_device_node_selection of this V1ResourceSliceSpec.  # noqa: E501

        PerDeviceNodeSelection defines whether the access from nodes to resources in the pool is set on the ResourceSlice level or on each device. If it is set to true, every device defined the ResourceSlice must specify this individually.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set.  # noqa: E501

        :return: The per_device_node_selection of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: bool
        """
        return self._per_device_node_selection

    @per_device_node_selection.setter
    def per_device_node_selection(self, per_device_node_selection):
        """Sets the per_device_node_selection of this V1ResourceSliceSpec.

        PerDeviceNodeSelection defines whether the access from nodes to resources in the pool is set on the ResourceSlice level or on each device. If it is set to true, every device defined the ResourceSlice must specify this individually.  Exactly one of NodeName, NodeSelector, AllNodes, and PerDeviceNodeSelection must be set.  # noqa: E501

        :param per_device_node_selection: The per_device_node_selection of this V1ResourceSliceSpec.  # noqa: E501
        :type per_device_node_selection: bool
        """

        self._per_device_node_selection = per_device_node_selection

    @property
    def pool(self):
        """Gets the pool of this V1ResourceSliceSpec.  # noqa: E501


        :return: The pool of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: V1ResourcePool
        """
        return self._pool

    @pool.setter
    def pool(self, pool):
        """Sets the pool of this V1ResourceSliceSpec.


        :param pool: The pool of this V1ResourceSliceSpec.  # noqa: E501
        :type pool: V1ResourcePool
        """
        if self.local_vars_configuration.client_side_validation and pool is None:  # noqa: E501
            raise ValueError("Invalid value for `pool`, must not be `None`")  # noqa: E501

        self._pool = pool

    @property
    def shared_counters(self):
        """Gets the shared_counters of this V1ResourceSliceSpec.  # noqa: E501

        SharedCounters defines a list of counter sets, each of which has a name and a list of counters available.  The names of the counter sets must be unique in the ResourcePool.  Only one of Devices and SharedCounters can be set in a ResourceSlice.  The maximum number of counter sets is 8.  # noqa: E501

        :return: The shared_counters of this V1ResourceSliceSpec.  # noqa: E501
        :rtype: list[V1CounterSet]
        """
        return self._shared_counters

    @shared_counters.setter
    def shared_counters(self, shared_counters):
        """Sets the shared_counters of this V1ResourceSliceSpec.

        SharedCounters defines a list of counter sets, each of which has a name and a list of counters available.  The names of the counter sets must be unique in the ResourcePool.  Only one of Devices and SharedCounters can be set in a ResourceSlice.  The maximum number of counter sets is 8.  # noqa: E501

        :param shared_counters: The shared_counters of this V1ResourceSliceSpec.  # noqa: E501
        :type shared_counters: list[V1CounterSet]
        """

        self._shared_counters = shared_counters

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ResourceSliceSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ResourceSliceSpec):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2024 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

import datetime
import re

import durationpy

# Initialize our RE statically, rather than compiling for every call. This has
# the downside that it'll get compiled at import time but that shouldn't
# really be a big deal.
reDuration = re.compile(r'^([0-9]{1,5}(h|m|s|ms)){1,4}$')

# maxDuration_ms is the maximum duration that GEP-2257 can support, in
# milliseconds.
maxDuration_ms = (((99999 * 3600) + (59 * 60) + 59) * 1_000) + 999


def parse_duration(duration) -> datetime.timedelta:
    """
    Parse GEP-2257 Duration format to a datetime.timedelta object.

    The GEP-2257 Duration format is a restricted form of the input to the Go
    time.ParseDuration function; specifically, it must match the regex
    "^([0-9]{1,5}(h|m|s|ms)){1,4}$".

    See https://gateway-api.sigs.k8s.io/geps/gep-2257/ for more details.

    Input: duration: string
    Returns: datetime.timedelta

    Raises: ValueError on invalid or unknown input

    Examples:
    >>> parse_duration("1h")
    datetime.timedelta(seconds=3600)
    >>> parse_duration("1m")
    datetime.timedelta(seconds=60)
    >>> parse_duration("1s")
    datetime.timedelta(seconds=1)
    >>> parse_duration("1ms")
    datetime.timedelta(microseconds=1000)
    >>> parse_duration("1h1m1s")
    datetime.timedelta(seconds=3661)
    >>> parse_duration("10s30m1h")
    datetime.timedelta(seconds=5410)

    Units are always required.
    >>> parse_duration("1")
    Traceback (most recent call last):
        ...
    ValueError: Invalid duration format: 1

    Floating-point and negative durations are not valid.
    >>> parse_duration("1.5m")
    Traceback (most recent call last):
        ...
    ValueError: Invalid duration format: 1.5m
    >>> parse_duration("-1m")
    Traceback (most recent call last):
        ...
    ValueError: Invalid duration format: -1m
    """

    if not reDuration.match(duration):
        raise ValueError("Invalid duration format: {}".format(duration))

    return durationpy.from_str(duration)


def format_duration(delta: datetime.timedelta) -> str:
    """
    Format a datetime.timedelta object to GEP-2257 Duration format.

    The GEP-2257 Duration format is a restricted form of the input to the Go
    time.ParseDuration function; specifically, it must match the regex
    "^([0-9]{1,5}(h|m|s|ms)){1,4}$".

    See https://gateway-api.sigs.k8s.io/geps/gep-2257/ for more details.

    Input: duration: datetime.timedelta

    Returns: string

    Raises: ValueError if the timedelta given cannot be expressed as a
    GEP-2257 Duration.

    Examples:
    >>> format_duration(datetime.timedelta(seconds=3600))
    '1h'
    >>> format_duration(datetime.timedelta(seconds=60))
    '1m'
    >>> format_duration(datetime.timedelta(seconds=1))
    '1s'
    >>> format_duration(datetime.timedelta(microseconds=1000))
    '1ms'
    >>> format_duration(datetime.timedelta(seconds=5410))
    '1h30m10s'

    The zero duration is always "0s".
    >>> format_duration(datetime.timedelta(0))
    '0s'

    Sub-millisecond precision is not allowed.
    >>> format_duration(datetime.timedelta(microseconds=100))
    Traceback (most recent call last):
        ...
    ValueError: Cannot express sub-millisecond precision in GEP-2257: 0:00:00.000100

    Negative durations are not allowed.
    >>> format_duration(datetime.timedelta(seconds=-1))
    Traceback (most recent call last):
        ...
    ValueError: Cannot express negative durations in GEP-2257: -1 day, 23:59:59
    """

    # Short-circuit if we have a zero delta.
    if delta == datetime.timedelta(0):
        return "0s"

    # Check range early.
    if delta < datetime.timedelta(0):
        raise ValueError("Cannot express negative durations in GEP-2257: {}".format(delta))

    if delta > datetime.timedelta(milliseconds=maxDuration_ms):
        raise ValueError(
            "Cannot express durations longer than 99999h59m59s999ms in GEP-2257: {}".format(delta))

    # durationpy.to_str() is happy to use floating-point seconds, which
    # GEP-2257 is _not_ happy with. So start by peeling off any microseconds
    # from our delta.
    delta_us = delta.microseconds

    if (delta_us % 1000) != 0:
        raise ValueError(
            "Cannot express sub-millisecond precision in GEP-2257: {}"
            .format(delta)
        )

    # After that, do the usual div & mod tree to take seconds and get hours,
    # minutes, and seconds from it.
    secs = int(delta.total_seconds())

    output: List[str] = []

    hours = secs // 3600
    if hours > 0:
        output.append(f"{hours}h")
        secs -= hours * 3600

    minutes = secs // 60
    if minutes > 0:
        output.append(f"{minutes}m")
        secs -= minutes * 60

    if secs > 0:
        output.append(f"{secs}s")

    if delta_us > 0:
        output.append(f"{delta_us // 1000}ms")

    return "".join(output)



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1CustomResourceDefinitionStatus(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'accepted_names': 'V1CustomResourceDefinitionNames',
        'conditions': 'list[V1CustomResourceDefinitionCondition]',
        'observed_generation': 'int',
        'stored_versions': 'list[str]'
    }

    attribute_map = {
        'accepted_names': 'acceptedNames',
        'conditions': 'conditions',
        'observed_generation': 'observedGeneration',
        'stored_versions': 'storedVersions'
    }

    def __init__(self, accepted_names=None, conditions=None, observed_generation=None, stored_versions=None, local_vars_configuration=None):  # noqa: E501
        """V1CustomResourceDefinitionStatus - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._accepted_names = None
        self._conditions = None
        self._observed_generation = None
        self._stored_versions = None
        self.discriminator = None

        if accepted_names is not None:
            self.accepted_names = accepted_names
        if conditions is not None:
            self.conditions = conditions
        if observed_generation is not None:
            self.observed_generation = observed_generation
        if stored_versions is not None:
            self.stored_versions = stored_versions

    @property
    def accepted_names(self):
        """Gets the accepted_names of this V1CustomResourceDefinitionStatus.  # noqa: E501


        :return: The accepted_names of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :rtype: V1CustomResourceDefinitionNames
        """
        return self._accepted_names

    @accepted_names.setter
    def accepted_names(self, accepted_names):
        """Sets the accepted_names of this V1CustomResourceDefinitionStatus.


        :param accepted_names: The accepted_names of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :type accepted_names: V1CustomResourceDefinitionNames
        """

        self._accepted_names = accepted_names

    @property
    def conditions(self):
        """Gets the conditions of this V1CustomResourceDefinitionStatus.  # noqa: E501

        conditions indicate state for particular aspects of a CustomResourceDefinition  # noqa: E501

        :return: The conditions of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :rtype: list[V1CustomResourceDefinitionCondition]
        """
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        """Sets the conditions of this V1CustomResourceDefinitionStatus.

        conditions indicate state for particular aspects of a CustomResourceDefinition  # noqa: E501

        :param conditions: The conditions of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :type conditions: list[V1CustomResourceDefinitionCondition]
        """

        self._conditions = conditions

    @property
    def observed_generation(self):
        """Gets the observed_generation of this V1CustomResourceDefinitionStatus.  # noqa: E501

        The generation observed by the CRD controller.  # noqa: E501

        :return: The observed_generation of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :rtype: int
        """
        return self._observed_generation

    @observed_generation.setter
    def observed_generation(self, observed_generation):
        """Sets the observed_generation of this V1CustomResourceDefinitionStatus.

        The generation observed by the CRD controller.  # noqa: E501

        :param observed_generation: The observed_generation of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :type observed_generation: int
        """

        self._observed_generation = observed_generation

    @property
    def stored_versions(self):
        """Gets the stored_versions of this V1CustomResourceDefinitionStatus.  # noqa: E501

        storedVersions lists all versions of CustomResources that were ever persisted. Tracking these versions allows a migration path for stored versions in etcd. The field is mutable so a migration controller can finish a migration to another version (ensuring no old objects are left in storage), and then remove the rest of the versions from this list. Versions may not be removed from `spec.versions` while they exist in this list.  # noqa: E501

        :return: The stored_versions of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :rtype: list[str]
        """
        return self._stored_versions

    @stored_versions.setter
    def stored_versions(self, stored_versions):
        """Sets the stored_versions of this V1CustomResourceDefinitionStatus.

        storedVersions lists all versions of CustomResources that were ever persisted. Tracking these versions allows a migration path for stored versions in etcd. The field is mutable so a migration controller can finish a migration to another version (ensuring no old objects are left in storage), and then remove the rest of the versions from this list. Versions may not be removed from `spec.versions` while they exist in this list.  # noqa: E501

        :param stored_versions: The stored_versions of this V1CustomResourceDefinitionStatus.  # noqa: E501
        :type stored_versions: list[str]
        """

        self._stored_versions = stored_versions

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1CustomResourceDefinitionStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1CustomResourceDefinitionStatus):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: release-1.36
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from kubernetes.client.configuration import Configuration


class V1alpha1ClusterTrustBundleSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'signer_name': 'str',
        'trust_bundle': 'str'
    }

    attribute_map = {
        'signer_name': 'signerName',
        'trust_bundle': 'trustBundle'
    }

    def __init__(self, signer_name=None, trust_bundle=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1ClusterTrustBundleSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._signer_name = None
        self._trust_bundle = None
        self.discriminator = None

        if signer_name is not None:
            self.signer_name = signer_name
        self.trust_bundle = trust_bundle

    @property
    def signer_name(self):
        """Gets the signer_name of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501

        signerName indicates the associated signer, if any.  In order to create or update a ClusterTrustBundle that sets signerName, you must have the following cluster-scoped permission: group=certificates.k8s.io resource=signers resourceName=<the signer name> verb=attest.  If signerName is not empty, then the ClusterTrustBundle object must be named with the signer name as a prefix (translating slashes to colons). For example, for the signer name `example.com/foo`, valid ClusterTrustBundle object names include `example.com:foo:abc` and `example.com:foo:v1`.  If signerName is empty, then the ClusterTrustBundle object's name must not have such a prefix.  List/watch requests for ClusterTrustBundles can filter on this field using a `spec.signerName=NAME` field selector.  # noqa: E501

        :return: The signer_name of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501
        :rtype: str
        """
        return self._signer_name

    @signer_name.setter
    def signer_name(self, signer_name):
        """Sets the signer_name of this V1alpha1ClusterTrustBundleSpec.

        signerName indicates the associated signer, if any.  In order to create or update a ClusterTrustBundle that sets signerName, you must have the following cluster-scoped permission: group=certificates.k8s.io resource=signers resourceName=<the signer name> verb=attest.  If signerName is not empty, then the ClusterTrustBundle object must be named with the signer name as a prefix (translating slashes to colons). For example, for the signer name `example.com/foo`, valid ClusterTrustBundle object names include `example.com:foo:abc` and `example.com:foo:v1`.  If signerName is empty, then the ClusterTrustBundle object's name must not have such a prefix.  List/watch requests for ClusterTrustBundles can filter on this field using a `spec.signerName=NAME` field selector.  # noqa: E501

        :param signer_name: The signer_name of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501
        :type signer_name: str
        """

        self._signer_name = signer_name

    @property
    def trust_bundle(self):
        """Gets the trust_bundle of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501

        trustBundle contains the individual X.509 trust anchors for this bundle, as PEM bundle of PEM-wrapped, DER-formatted X.509 certificates.  The data must consist only of PEM certificate blocks that parse as valid X.509 certificates.  Each certificate must include a basic constraints extension with the CA bit set.  The API server will reject objects that contain duplicate certificates, or that use PEM block headers.  Users of ClusterTrustBundles, including Kubelet, are free to reorder and deduplicate certificate blocks in this file according to their own logic, as well as to drop PEM block headers and inter-block data.  # noqa: E501

        :return: The trust_bundle of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501
        :rtype: str
        """
        return self._trust_bundle

    @trust_bundle.setter
    def trust_bundle(self, trust_bundle):
        """Sets the trust_bundle of this V1alpha1ClusterTrustBundleSpec.

        trustBundle contains the individual X.509 trust anchors for this bundle, as PEM bundle of PEM-wrapped, DER-formatted X.509 certificates.  The data must consist only of PEM certificate blocks that parse as valid X.509 certificates.  Each certificate must include a basic constraints extension with the CA bit set.  The API server will reject objects that contain duplicate certificates, or that use PEM block headers.  Users of ClusterTrustBundles, including Kubelet, are free to reorder and deduplicate certificate blocks in this file according to their own logic, as well as to drop PEM block headers and inter-block data.  # noqa: E501

        :param trust_bundle: The trust_bundle of this V1alpha1ClusterTrustBundleSpec.  # noqa: E501
        :type trust_bundle: str
        """
        if self.local_vars_configuration.client_side_validation and trust_bundle is None:  # noqa: E501
            raise ValueError("Invalid value for `trust_bundle`, must not be `None`")  # noqa: E501

        self._trust_bundle = trust_bundle

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1ClusterTrustBundleSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1ClusterTrustBundleSpec):
            return True

        return self.to_dict() != other.to_dict()
